from abc import ABC
import os

import yaml
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches

from .com import total_com
from .geometry import fundamental_geometry_plot_data
from .conversions import _convert_principal_to_benchmark


def _com_symbol(ax, center, radius, color='b', label=None):
    """Returns matplotlib axis with center of mass symbol and possibly a label
    added."""
    c = patches.Circle(center, radius=radius, fill=False)
    w1 = patches.Wedge(center, radius, 0., 90., color=color, ec=None,
                       alpha=0.5)
    w2 = patches.Wedge(center, radius, 180., 270., color=color, ec=None,
                       alpha=0.5)
    ax.add_patch(w1)
    ax.add_patch(w2)
    ax.add_patch(c)

    if label is not None:
        ax.text(center[0] + radius, center[1] + radius, label)

    return ax


def _plot_geometry_from_benchmark(ax, parameters, show_steer_axis=True):
    """Returns a matplotlib axis with the basic bicycle geometry plotted."""

    p = parameters

    # plot the ground line
    ax.axhline(0.0, color='black')

    # plot the rear wheel
    c = patches.Circle((0., -p['rR']), radius=p['rR'],
                       linewidth=2, fill=False)
    ax.add_patch(c)

    # plot the front wheel
    c = patches.Circle((p['w'], -p['rF']), radius=p['rF'],
                       linewidth=2, fill=False)
    ax.add_patch(c)

    # plot the fundamental bike (Moore2012 depiction)
    deex, deez = fundamental_geometry_plot_data(p)
    ax.plot(deex, deez, color='k', linewidth=2)

    if show_steer_axis:
        # plot the steer axis
        dx3 = deex[2] - deez[2] * (deex[2] - deex[1]) / (-deez[1] + deez[2])
        ax.plot([deex[2], dx3], [deez[2], 0.], 'k--')

    return ax


class ParameterSet(ABC):
    """A parameter set is a collection of constants with associated floating
    point values that are present in a set of differential algebraic equations
    that represent a multibody bicycle model. These pairs are typically defined
    in a specific academic article, dissertation, book chapter, or section and
    subclasses should be named in a way that ties them to that written work.
    Parameter sets must be named with this pattern ``NameOfMySetParameterSet``
    where ``NameOfMySet`` is a unique name other than something that includes
    ``ParameterSet``.

    A parameter set, or a subset of the parameters, can be used with multiple
    different models. A parameter set is associated with a particular
    parameterization of one or more models. Parameter sets can be converted to
    equivalent parameter sets, but only by assuming a particular model
    configuration. The obvious configuration for a bicycle model is the
    upright, zero steer state. But if, for example, a rider configuration is
    included, then some nominal configuration would need to be defined for
    conversion consistency.

    Each parameter set should have a unique set of ASCII strings that represent
    the constants in a model.

    Attributes
    ==========
    par_strings : dictionary
        Maps ASCII strings to their LaTeX string.

    """
    par_strings = {
        'aR': r'a_R',
        'beta': r'\beta',
    }

    def _check_parameters(self, parameters):
        """Ensures that each parameter in par_strings is present in parameters
        and that the values are floats."""
        for k, _ in self.par_strings.items():
            if k not in parameters.keys():
                msg = '{} is missing from the provided parameter dictionary.'
                raise ValueError(msg.format(k))
            if not isinstance(parameters[k], float):
                msg = '{} is not a valid value for parameter {}'
                raise ValueError(msg.format(parameters[k], k))

    def __init__(self, par_dict):
        self._check_parameters(par_dict)
        cls_name = self.__class__.__name__
        self.parameterization = cls_name.split('ParameterSet')[0]

    @classmethod
    def _from_yaml(cls, path_to_file):
        """Instantiates a ParameterSet from a file."""
        # TODO : Wouldn't this have to dynamically create the class so that the
        # class name is derived from a specifier in the yaml file?
        with open(os.path.join(path_to_file), 'r') as f:
            yaml_data = yaml.load(f, Loader=yaml.FullLoader)
        return cls(yaml_data['values'])

    def _generate_body_colors(self):
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']
        self.body_colors = {}
        for i, b in enumerate(self.body_labels):
            self.body_colors[b] = colors[i]

    def _invert_yaxis(self, ax):
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
        return ax

    def _finalize_plot(self, ax):
        ax = self._invert_yaxis(ax)
        ax.set_aspect('equal')
        ax.autoscale()  # needed for the plots with only patches
        ax.set_xlabel(r'$x$ [m]')
        ax.set_ylabel(r'$z$ [m]')

    def to_parameterization(self, name):
        """Returns a specific parameter set based on the provided
        parameterization name.

        Parameters
        ==========
        name : string
            The name of the parameterization. These should correspond to a
            subclass of a ``ParameterSet`` and the name will be the string that
            precedes "ParameterSet". For example, the parameterization name of
            ``Meijaard2007ParameterSet`` is ``Meijaard2007``.

        Returns
        =======
        ParmeterSet
            If a different parameterization is requested and this class can
            convert itself, it will return a new parameter set of the correct
            parameterization.

        """
        if name == self.parameterization:
            return self
        else:
            msg = '{} is not an avaialble parameter set'
            raise ValueError(msg.format(name))

    def to_yaml(self, fname):
        """Writes parameters to file in the YAML format.

        Parameters
        ==========
        fname : string
            Path to file.

        """
        data = {
            'description': '',
            'parameterization': self.parameterization,
            'parameters': '',
            'rider': None,
            'values': self.parameters,
        }
        with open(fname, 'w') as f:
            yaml.dump(data, f)

    def to_ini(self, fname):
        """Writes parameters to file in the INI format. Metadata is not
        included.

        Parameters
        ==========
        fname : string
            Path to file.

        """
        text = ""
        for k, v in self.parameters.items():
            text += "{}={:1.16f}\n".format(k, v)
        with open(fname, 'w') as f:
            f.write(text)


class Meijaard2007ParameterSet(ParameterSet):
    """Represents the parameters of the benchmark bicycle presented in
    [Meijaard2007]_.

    The four bodies are:

    - B: rear frame + rigid rider
    - F: front wheel
    - H: front frame (fork & handlebars)
    - R: rear wheel

    Parameters
    ==========
    parameters : dictionary
        A dictionary mapping variable names to values that contains the
        following keys:

        - ``IBxx`` : x moment of inertia of the frame/rider [kg*m**2]
        - ``IBxz`` : xz product of inertia of the frame/rider [kg*m**2]
        - ``IBzz`` : z moment of inertia of the frame/rider [kg*m**2]
        - ``IFxx`` : x moment of inertia of the front wheel [kg*m**2]
        - ``IFyy`` : y moment of inertia of the front wheel [kg*m**2]
        - ``IHxx`` : x moment of inertia of the handlebar/fork [kg*m**2]
        - ``IHxz`` : xz product of inertia of the handlebar/fork [kg*m**2]
        - ``IHzz`` : z moment of inertia of the handlebar/fork [kg*m**2]
        - ``IRxx`` : x moment of inertia of the rear wheel [kg*m**2]
        - ``IRyy`` : y moment of inertia of the rear wheel [kg*m**2]
        - ``c`` : trail [m]
        - ``g`` : acceleration due to gravity [m/s**2]
        - ``lam`` : steer axis tilt [rad]
        - ``mB`` : frame/rider mass [kg]
        - ``mF`` : front wheel mass [kg]
        - ``mH`` : handlebar/fork assembly mass [kg]
        - ``mR`` : rear wheel mass [kg]
        - ``rF`` : front wheel radius [m]
        - ``rR`` : rear wheel radius [m]
        - ``w`` : wheelbase [m]
        - ``xB`` : x distance to the frame/rider center of mass [m]
        - ``xH`` : x distance to the frame/rider center of mass [m]
        - ``zB`` : z distance to the frame/rider center of mass [m]
        - ``zH`` : z distance to the frame/rider center of mass [m]

    includes_rider : boolean
        True if body B is the combined rear frame and rider in terms of
        mass and inertia values.

    Attributes
    ==========
    par_strings : dictionary
        Maps ASCII strings to their LaTeX string.
    body_labels : list of strings
        Single capital letters that correspond to the four rigid bodies in the
        model.

    References
    ==========

    .. [Meijaard2007] Meijaard J.P, Papadopoulos Jim M, Ruina Andy and Schwab
       A.L, 2007, Linearized dynamics equations for the balance and steer of a
       bicycle: a benchmark and review, Proc. R. Soc. A., 463:1955–1982
       http://doi.org/10.1098/rspa.2007.1857

    """

    # maps "Python" string to LaTeX version
    par_strings = {
        'IBxx': r'I_{Bxx}',
        'IBxz': r'I_{Bxz}',
        'IByy': r'I_{Byy}',
        'IBzz': r'I_{Bzz}',
        'IFxx': r'I_{Fxx}',
        'IFyy': r'I_{Fyy}',
        'IHxx': r'I_{Hxx}',
        'IHxz': r'I_{Hxz}',
        'IHyy': r'I_{Hyy}',
        'IHzz': r'I_{Hzz}',
        'IRxx': r'I_{Rxx}',
        'IRyy': r'I_{Ryy}',
        'c': r'c',
        'g': r'g',
        'lam': r'\lambda',
        'mB': r'm_B',
        'mF': r'm_F',
        'mH': r'm_H',
        'mR': r'm_R',
        'rF': r'r_F',
        'rR': r'r_R',
        'v': r'v',
        'w': r'w',
        'xB': r'x_B',
        'xH': r'x_H',
        'zB': r'z_B',
        'zH': r'z_H',
    }

    body_labels = ['B', 'F', 'H', 'R']

    def __init__(self, parameters, includes_rider):
        super().__init__(parameters)
        self.parameters = parameters
        self.includes_rider = includes_rider
        self._generate_body_colors()

    @classmethod
    def _from_yaml(cls, path_to_file):

        with open(os.path.join(path_to_file), 'r') as f:
            yaml_data = yaml.load(f, Loader=yaml.FullLoader)
            benchmark_par = yaml_data['values']
            rider = yaml_data['rider']

        return cls(benchmark_par, rider)

    def _calc_derived_params(self):
        # These parameters are needed but are not specified by the user.
        p = self.parameters

        pext = {}
        pext['IFxz'] = 0.0
        pext['IFzz'] = p['IFxx']
        pext['IRxz'] = 0.0
        pext['IRzz'] = p['IRxx']
        pext['xF'] = p['w']
        pext['xR'] = 0.0
        pext['yB'] = 0.0
        pext['yF'] = 0.0
        pext['yH'] = 0.0
        pext['yR'] = 0.0
        pext['zF'] = -p['rF']
        pext['zR'] = -p['rR']

        return pext

    def form_mass_center_vector(self, body):
        """Returns an array representing the 3D vector to the mass center of
        the body from the origin at the rear wheel contact point.

        Parameters
        ==========
        body : string
            One of 'B', 'F', 'H', 'R'.

        Returns
        =======
        ndarray, shape(3,)
            A vector containing the X, Y, and X coordinates of the mass center
            of the body.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> p.form_mass_center_vector('B')
        array([ 0.28909943,  0.        , -1.04029228])

        """

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        x = p['x{}'.format(body)]
        y = p['y{}'.format(body)]
        z = p['z{}'.format(body)]

        return np.array([x, y, z])

    def mass_center_of(self, *bodies):
        """Returns the vector locating the center of mass of the collection of
        bodies.

        Parameters
        ==========
        bodies : iterable of strings
            One or more of the ``body_labels``.

        Returns
        =======
        com : ndarray, shape(3,)
            Vector locating the center of mass of the bodies givien in
            ``bodies``.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> p.mass_center_of('B', 'H')
        array([ 0.31096918,  0.        , -1.02923892])

        """
        if len(bodies) == 1:
            return self.form_mass_center_vector(bodies[0])
        else:
            coordinates = []
            masses = []
            for body in bodies:
                masses.append(self.parameters['m{}'.format(body)])
                coordinates.append(
                    self.form_mass_center_vector(body))

            coordinates = np.array(coordinates).T

            mass, com = total_com(coordinates, masses)

            return com

    def form_inertia_tensor(self, body):
        """Returns the inertia tensor with respect to the global coordinate
        system and the body's mass center.

        Parameters
        ==========
        body : string
            One of the ``body_labels``.

        Returns
        =======
        inertia_tensor : ndarray, shape(3, 3)
            Inertia tensor of the body with respect to the body's mass center
            and the model's coordinate system.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> p.form_inertia_tensor('H')
        array([[ 0.25337959,  0.        , -0.07204524],
               [ 0.        ,  0.24613881,  0.        ],
               [-0.07204524,  0.        ,  0.09557708]])

        """

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        Ixx = p['I{}xx'.format(body)]
        Ixz = p['I{}xz'.format(body)]
        Iyy = p['I{}yy'.format(body)]
        Izz = p['I{}zz'.format(body)]

        inertia_tensor = np.array([[Ixx, 0.0, Ixz],
                                   [0.0, Iyy, 0.0],
                                   [Ixz, 0.0, Izz]])

        return inertia_tensor

    def plot_geometry(self, show_steer_axis=True, ax=None):
        """Returns a matplotlib axes with a simple drawing of the bicycle's
        geometry.

        Parameters
        ==========
        show_steer_axis : boolean
            If true, a dotted line will be plotted along the steer axis from
            the front wheel center to the ground.
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_geometry()

        """

        p = self.parameters

        if ax is None:
            fig, ax = plt.subplots()

        ax = _plot_geometry_from_benchmark(ax, p,
                                           show_steer_axis=show_steer_axis)

        self._finalize_plot(ax)

        return ax

    def plot_mass_centers(self, bodies=None, ax=None):
        """Returns a matplotlib axes with mass center indicators for each body.

        Parameters
        ==========
        bodies: list of strings, optional
            A subset of the strings present in the class attribute
            ``body_labels``.
        ax: matplotlib Axes, optional
            An axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_mass_centers()

        """

        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = self.body_labels

        for b in bodies:
            ax = self.plot_body_mass_center(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_body_mass_center(self, body, ax=None):
        """Returns a matplotlib axes with a mass center symbol for the
        specified body to the plot.

        Parameters
        ==========
        body : string
            The body string: ``F``, ``H``, ``B``, or ``R``.
        ax : SubplotAxes, optional
            Axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_body_mass_center('B')

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        # mass center
        x = p['x{}'.format(body)]
        z = p['z{}'.format(body)]
        radius = p['w'] / 10
        ax = _com_symbol(ax, (x, z), radius, color=self.body_colors[body],
                         label=body)
        self._finalize_plot(ax)

        return ax

    def _planar_principal_radii_of_gyration(self, body):

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        IH = self.form_inertia_tensor(body)
        # remove the Y row and column
        IH_xz_plane = np.delete(np.delete(IH, 1, axis=1), 1, axis=0)

        evals, evecs = np.linalg.eig(IH_xz_plane)

        idxs = np.argsort(evals)

        # NOTE : min is first entry, max is second entry
        evals = evals[idxs]
        evecs = evecs[:, idxs]

        # NOTE : The negative sign on the z value ensures the sign of the
        # rotation about the Y axis is correct, i.e. arctan2 is thinking you
        # are rotating in XY plane about Z.
        angle_to_max = np.arctan2(-evecs[1, 1], evecs[0, 1])

        kmin = np.sqrt(evals[0] / p['m{}'.format(body)])
        kmax = np.sqrt(evals[1] / p['m{}'.format(body)])
        kyy = np.sqrt(p['I{}yy'.format(body)] / p['m{}'.format(body)])

        return kmax, kmin, kyy, angle_to_max

    def plot_body_principal_radii_of_gyration(self, body, ax=None):
        """Returns a matplotlib axes with lines and a circle that indicate the
        principal radii of gyration of the specified body.

        Parameters
        ==========
        body : string
            One of the ``body_labels``.
        ax : SubplotAxes, optional
            Axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_body_principal_radii_of_gyration('B')

        """
        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        x = p['x{}'.format(body)]
        z = p['z{}'.format(body)]
        kaa, kbb, kyy, alpha = self._planar_principal_radii_of_gyration(body)

        c = patches.Circle((x, z), radius=kyy, fill=False,
                           color=self.body_colors[body], linestyle='--')
        ax.add_patch(c)

        # NOTE : -alpha is required because we are mapping the xz axes to a new
        # planar drawing grid which is x and y with z pointing out of the
        # screen
        ax.plot([x - kbb*np.cos(-alpha), x + kbb*np.cos(-alpha)],
                [z - kbb*np.sin(-alpha), z + kbb*np.sin(-alpha)],
                color=self.body_colors[body], linestyle="--")

        ax.plot([x - kaa*np.cos(-alpha - np.pi/2),
                 x + kaa*np.cos(-alpha - np.pi/2)],
                [z - kaa*np.sin(-alpha - np.pi/2),
                 z + kaa*np.sin(-alpha - np.pi/2)],
                color=self.body_colors[body], linestyle="--")

        self._finalize_plot(ax)

        return ax

    def plot_principal_radii_of_gyration(self, bodies=None, ax=None):
        """Returns a matplotlib axis with principal radii of all bodies
        shown.

        Parameters
        ==========
        bodies: list of strings, optional
            A subset of the strings present in the class attribute
            ``body_labels``.
        ax: matplotlib Axes, optional
            An axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_principal_radii_of_gyration()

        """
        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = self.body_labels

        for b in bodies:
            ax = self.plot_body_principal_radii_of_gyration(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_body_principal_inertia_ellipsoid(self, body, ax=None):
        """Returns a matplotlib axes with an ellipse that respresnts the XZ
        plane view of a constant density ellipsoid which has the same principal
        moments and axes of inertia as the body.

        Parameters
        ==========
        body : string
            One of the ``body_labels``.
        ax : SubplotAxes, optional
            Axes to plot on.


        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_body_principal_inertia_ellipsoid('H')

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        kaa, kbb, kyy, alpha = self._planar_principal_radii_of_gyration(body)

        width = np.sqrt(5/2*(-kaa**2 + kyy**2 + kbb**2))
        height = np.sqrt(5/2*(kaa**2 + kyy**2 - kbb**2))

        ellipse = patches.Ellipse((p['x{}'.format(body)], p['z{}'.format(body)]),
                                  width, height,
                                  angle=-np.rad2deg(alpha), fill=False,
                                  color=self.body_colors[body])
        ax.add_patch(ellipse)

        self._finalize_plot(ax)

        return ax

    def plot_principal_inertia_ellipsoids(self, bodies=None, ax=None):
        """Returns a Matplotlib axes with 2D representations of 3D solid
        uniform ellipsoids that have the same inertia as the body.

        Parameters
        ==========
        bodies: list of strings, optional
            A subset of the strings present in the class attribute
            ``body_labels``.
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_principal_inertia_ellipsoids()

        """
        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = self.body_labels

        for b in bodies:
            ax = self.plot_body_principal_inertia_ellipsoid(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_all(self, ax=None):
        """Returns matplotlib axes with the geometry and inertial
        representations of all bodies of the bicycle parameter set.

        Parameters
        ==========
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           p.plot_all()

        """
        if ax is None:
            fig, ax = plt.subplots()

        ax = self.plot_principal_radii_of_gyration(ax=ax)
        ax = self.plot_principal_inertia_ellipsoids(ax=ax)
        ax = self.plot_geometry(ax=ax)
        ax = self.plot_mass_centers(ax=ax)

        return ax


class Moore2019ParameterSet(ParameterSet):
    """Represents the parameters of the bicycle parameterization presented in
    [1]_.

    The four bodies are:

    - D: rear frame
    - F: front wheel
    - H: front frame (fork & handlebars)
    - P: rigid rider
    - R: rear wheel

    Parameters
    ==========
    parameters : dictionary
        A dictionary mapping variable names to values.

    Attributes
    ==========
    par_strings : dictionary
        Maps ASCII strings to their LaTeX string.
    body_labels : list of strings
        Single capital letters that correspond to the five rigid bodies in the
        model.

    References
    ==========

    .. [1] Moore, Jason K.; Hubbard, Mont (2019): Expanded Optimization for
       Discovering Optimal Lateral Handling Bicycles. Proceedings of Bicycle
       and Motorcycle Dynamics 2019: A Symposium on the Dynamics and Control of
       Single Track Vehicles https://doi.org/10.6084/m9.figshare.9942938.v1

    """

    non_min_par_strings = {
        'alphaF': r'\alpha_F',
        'alphaR': r'\alpha_R',
        'kFbb': r'k_{Fbb}',
        'kRbb': r'k_{Rbb}',
        'yD': r'y_D',
        'yF': r'y_F',
        'yH': r'y-H',
        'yP': r'y_P',
        'yR': r'y_R',
        'zR': r'z_R',
        'zF': r'z_F',
        'xR': r'x_R',
        'xF': r'x_F',
    }
    # maps "Python" string to LaTeX version
    par_strings = {
        'alphaD': r'\alpha_D',
        'alphaH': r'\alpha_H',
        'alphaP': r'\alpha_P',
        'c': r'c',
        'g': r'g',
        'kDaa': r'k_{Daa}',
        'kDbb': r'k_{Dbb}',
        'kDyy': r'k_{Dyy}',
        'kFaa': r'k_{Faa}',
        'kFyy': r'k_{Fyy}',
        'kHaa': r'k_{Haa}',
        'kHbb': r'k_{Hbb}',
        'kHyy': r'k_{Hyy}',
        'kPaa': r'k_{Paa}',
        'kPbb': r'k_{Pbb}',
        'kPyy': r'k_{Pyy}',
        'kRaa': r'k_{Raa}',
        'kRyy': r'k_{Ryy}',
        'lP': r'l_P',
        'lam': r'\lambda',
        'mD': r'm_D',
        'mF': r'm_F',
        'mH': r'm_H',
        'mP': r'm_B',
        'mR': r'm_R',
        'rF': r'r_F',
        'rR': r'r_R',
        'v': r'v',
        'w': r'w',
        'wP': r'w_P',
        'xD': r'x_D',
        'xH': r'x_H',
        'xP': r'x_P',
        'zD': r'z_D',
        'zH': r'z-H',
        'zP': r'z_P',
    }

    body_labels = ['D', 'F', 'H', 'P', 'R']

    def __init__(self, parameters):
        super().__init__(parameters)
        self.parameters = parameters
        self._generate_body_colors()

    def _calc_derived_params(self):
        p = self.parameters

        pext = {}
        pext['alphaF'] = 0.0
        pext['alphaR'] = 0.0
        pext['yD'] = 0.0
        pext['yP'] = 0.0
        pext['yH'] = 0.0
        pext['yR'] = 0.0
        pext['yF'] = 0.0
        pext['xR'] = 0.0
        pext['xF'] = p['w']
        pext['zR'] = -p['rR']
        pext['zF'] = -p['rF']
        pext['kRbb'] = p['kRaa']
        pext['kFbb'] = p['kFaa']

        return pext

    def to_parameterization(self, name):
        """Returns a specific parameter set based on the provided
        parameterization name.

        Parameters
        ==========
        name : string
            The name of the parameterization. These should correspond to a
            subclass of a ``ParameterSet`` and the name will be the string that
            precedes "ParameterSet". For example, the parameterization name of
            ``Meijaard2007ParameterSet`` is ``Meijaard2007``.

        Returns
        =======
        ParmeterSet
            If a different parameterization is requested and this class can
            convert itself, it will return a new parameter set of the correct
            parameterization.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import moore2019_browser_jason
        >>> from bicycleparameters.parameter_sets import Moore2019ParameterSet
        >>> moore_set = Moore2019ParameterSet(moore2019_browser_jason)
        >>> moore_set.mass_center_of('P', 'D')
        array([ 0.31156277,  0.        , -1.03972442])
        >>> meijaard_set = moore_set.to_parameterization('Meijaard2007')
        >>> meijaard_set.mass_center_of('B')
        array([ 0.31156277,  0.        , -1.03972442])

        """
        if name == self.parameterization:
            return self
        elif name == 'Meijaard2007':
            b = _convert_principal_to_benchmark(self.parameters)
            # Moore2019 always includes the rider
            return Meijaard2007ParameterSet(b, True)
        else:
            msg = '{} is not an avaialble parameter set'
            raise ValueError(msg.format(name))

    def plot_geometry(self, show_steer_axis=True, ax=None):
        """Returns a matplotlib axes with the simplest drawing of the bicycle's
        geometry.

        Parameters
        ==========
        show_steer_axis : boolean
            If true, a dotted line will be plotted along the steer axis from
            the front wheel center to the ground.
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_geometry()

        """

        p = self.parameters

        if ax is None:
            fig, ax = plt.subplots()

        ax = _plot_geometry_from_benchmark(ax, p,
                                           show_steer_axis=show_steer_axis)

        self._finalize_plot(ax)

        return ax

    def plot_person_diamond(self, show_cross=False, ax=None):
        """Plots a diamond that represents the approximate person's physical
        extents.

        Parameters
        ==========
        show_cross : boolean, optional
            Plots a cross in the diamond that spans opposite vertices.
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_person_diamond()

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters

        x_offset = np.array([
            p['wP']/2*np.cos(-p['alphaP']),
            -p['lP']/2*np.cos(-p['alphaP'] + np.pi/2),
            -p['wP']/2*np.cos(-p['alphaP']),
            p['lP']/2*np.cos(-p['alphaP'] + np.pi/2),
            p['wP']/2*np.cos(-p['alphaP'])])
        x = p['xP'] + x_offset

        z_offset = np.array([
            p['wP']/2*np.sin(-p['alphaP']),
            -p['lP']/2*np.sin(-p['alphaP'] + np.pi/2),
            -p['wP']/2*np.sin(-p['alphaP']),
            p['lP']/2*np.sin(-p['alphaP'] + np.pi/2),
            p['wP']/2*np.sin(-p['alphaP'])])
        z = p['zP'] + z_offset

        ax.plot(x, z, color=self.body_colors['P'])

        if show_cross:
            ax.plot([p['xP'] - p['wP']/2*np.cos(-p['alphaP']),
                     p['xP'] + p['wP']/2*np.cos(-p['alphaP'])],
                    [p['zP'] - p['wP']/2*np.sin(-p['alphaP']),
                     p['zP'] + p['wP']/2*np.sin(-p['alphaP'])],
                    color='black', linewidth=2)

            ax.plot([p['xP'] - p['lP']/2*np.cos(-p['alphaP'] + np.pi/2),
                     p['xP'] + p['lP']/2*np.cos(-p['alphaP'] + np.pi/2)],
                    [p['zP'] - p['lP']/2*np.sin(-p['alphaP'] + np.pi/2),
                     p['zP'] + p['lP']/2*np.sin(-p['alphaP'] + np.pi/2)],
                    color='black', linewidth=2)

        self._finalize_plot(ax)

        return ax

    def plot_mass_centers(self, bodies=None, ax=None):
        """Returns a matplotlib axes with a mass center symbols for the
        specified bodies to the plot.

        Parameters
        ==========
        bodies: list of strings, optional
            A subset of the strings present in the class attribute
            ``body_labels``.
        ax: matplotlib Axes, optional
            An axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_mass_centers()

        """

        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = self.body_labels

        for b in bodies:
            ax = self.plot_body_mass_center(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_body_mass_center(self, body, ax=None):
        """Returns a matplotlib axes with a mass center symbol for the
        specified body to the plot.

        Parameters
        ==========
        body : string
            The body string: D, F, H, P, or R
        ax : SubplotAxes, optional
            Axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_body_mass_center('D')

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        # mass center
        x = p['x{}'.format(body)]
        z = p['z{}'.format(body)]
        radius = max(p['w'], p['lP']) / 30
        ax = _com_symbol(ax, (x, z), radius,
                         color=self.body_colors[body], label=body)

        self._finalize_plot(ax)

        return ax

    def plot_principal_radii_of_gyration(self, bodies=None, ax=None):
        """Returns a matplotlib axes with lines and a circle that indicate the
        principal radii of gyration for all five bodies.

        Parameters
        ==========
        bodies : list of strings, optional
            Either ['D', 'F', 'H', 'P', 'R'] or a subset thereof.
        ax : AxesSubplot, optional
            An axes to draw on, otherwise one is created.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_principal_radii_of_gyration()

        """

        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = ['D', 'F', 'H', 'P', 'R']

        for b in bodies:
            ax = self.plot_body_principal_radii_of_gyration(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_body_principal_radii_of_gyration(self, body, ax=None):
        """Returns a matplotlib axes with lines and a circle that indicate the
        principal radii of gyration of the specified body.

        Parameters
        ==========
        body : string
            The body string: D, F, H, P, or R
        ax : SubplotAxes, optional
            Axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_body_principal_radii_of_gyration('P')

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        x = p['x{}'.format(body)]
        z = p['z{}'.format(body)]
        kyy = p['k{}yy'.format(body)]
        kaa = p['k{}aa'.format(body)]
        kbb = p['k{}bb'.format(body)]
        alpha = p['alpha{}'.format(body)]  # angle between x and aa about y

        linestyle = '--'

        c = patches.Circle((x, z), radius=kyy, fill=False,
                           color=self.body_colors[body], linestyle=linestyle)
        ax.add_patch(c)

        # NOTE : -alpha is required because we are mapping the xz axes to a new
        # planar drawing grid which is x and y with z pointing out of the
        # screen
        ax.plot([x - kbb*np.cos(-alpha), x + kbb*np.cos(-alpha)],
                [z - kbb*np.sin(-alpha), z + kbb*np.sin(-alpha)],
                color=self.body_colors[body], linestyle=linestyle)

        ax.plot([x - kaa*np.cos(-alpha - np.pi/2),
                 x + kaa*np.cos(-alpha - np.pi/2)],
                [z - kaa*np.sin(-alpha - np.pi/2),
                 z + kaa*np.sin(-alpha - np.pi/2)],
                color=self.body_colors[body], linestyle=linestyle)

        self._finalize_plot(ax)

        return ax

    def plot_body_principal_inertia_ellipsoid(self, body, ax=None):
        """Returns a matplotlib axes with an ellipse that respresnts the XZ
        plane view of a constant density ellipsoid which has the same principal
        moments and axes of inertia as the body.

        Parameters
        ==========
        body : string
            The body string: D, F, H, P, or R
        ax : SubplotAxes, optional
            Axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_body_principal_inertia_ellipsoid('P')

        """

        if ax is None:
            fig, ax = plt.subplots()

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        kaa = p['k{}aa'.format(body)]
        kbb = p['k{}bb'.format(body)]
        kyy = p['k{}yy'.format(body)]
        alpha = p['alpha{}'.format(body)]

        width = np.sqrt(5/2*(-kaa**2 + kyy**2 + kbb**2))
        height = np.sqrt(5/2*(kaa**2 + kyy**2 - kbb**2))

        ellipse = patches.Ellipse((p['x{}'.format(body)],
                                   p['z{}'.format(body)]), width, height,
                                  angle=-np.rad2deg(alpha), fill=False,
                                  color=self.body_colors[body])
        ax.add_patch(ellipse)

        self._finalize_plot(ax)

        return ax

    def plot_principal_inertia_ellipsoids(self, bodies=None, ax=None):
        """Returns a Matplotlib axes with 2D representations of 3D solid
        uniform ellipsoids that have the same inertia as the body.

        Parameters
        ==========
        bodies: list of strings, optional
            A subset of the strings present in the class attribute
            ``body_labels``.
        ax: matplotlib Axes, optional
            An axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_principal_inertia_ellipsoids()

        """
        if ax is None:
            fig, ax = plt.subplots()

        if bodies is None:
            bodies = self.body_labels

        for b in bodies:
            ax = self.plot_body_principal_inertia_ellipsoid(b, ax=ax)

        self._finalize_plot(ax)

        return ax

    def plot_all(self, ax=None):
        """Returns matplotlib axes with the geometry and inertial
        representations of all bodies of the bicycle parameter set.

        Parameters
        ==========
        ax: matplotlib Axes, optional
            An axes to plot on.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           from bicycleparameters.parameter_dicts import moore2019_browser_jason
           from bicycleparameters.parameter_sets import Moore2019ParameterSet
           p = Moore2019ParameterSet(moore2019_browser_jason)
           p.plot_all()

        """

        if ax is None:
            fig, ax = plt.subplots()

        ax = self.plot_principal_radii_of_gyration(ax=ax)
        ax = self.plot_principal_inertia_ellipsoids(ax=ax)
        ax = self.plot_person_diamond(ax=ax)
        ax = self.plot_geometry(ax=ax)
        ax = self.plot_mass_centers(ax=ax)

        self._finalize_plot(ax)

        return ax

    def mass_center_of(self, *bodies):
        """Returns the vector locating the center of mass of the collection of
        bodies.

        Parameters
        ==========
        bodies : iterable of strings
            Subset from ``body_labels``.

        Returns
        =======
        com : ndarray, shape(3,)
            Vector locating the center of mass of the bodies given in
            ``bodies``.

        """
        if len(bodies) == 1:
            return self.form_mass_center_vector(bodies[0])
        else:
            coordinates = []
            masses = []
            for body in bodies:
                masses.append(self.parameters['m{}'.format(body)])
                coordinates.append(
                    self.form_mass_center_vector(body))

            coordinates = np.array(coordinates).T

            mass, com = total_com(coordinates, masses)

            return com

    def form_mass_center_vector(self, body):
        """Returns an array representing the vector to the mass center of the
        body.

        Parameters
        ==========
        body : string
            One of 'P', 'D', 'F', 'H', 'R'.

        Returns
        =======
        ndarray, shape(3,)
            A vector containing the X, Y, and X coordinates of the mass center
            of the body.

        """

        p = self.parameters.copy()
        p.update(self._calc_derived_params())

        x = p['x{}'.format(body)]
        y = p['y{}'.format(body)]
        z = p['z{}'.format(body)]

        return np.array([x, y, z])



#!/usr/bin/env python

# builtin modules
import os

# dependencies
from dtk import control
from matplotlib.patches import Ellipse, Wedge
from uncertainties import unumpy
import matplotlib.pyplot as plt
import numpy as np

try:
    import plotly
except ImportError:
    px = None
    go = None
else:
    del plotly
    import plotly.express as px
    import plotly.graph_objects as go

# local module imports
from . import bicycle
from . import inertia
from . import com
from . import io
from . import geometry
from . import period
from . import rider
from . import plot

GOLDEN_RATIO = (1.0 + np.sqrt(5.0))/2.0


class Bicycle(object):
    """
    An object for a bicycle. A bicycle has parameters and can have a rider
    attached to it. That's about it for now.

    """

    def __new__(cls, bicycleName, pathToData='.', forceRawCalc=False,
                forcePeriodCalc=False):
        '''Returns a NoneType object if there is no directory for the
        bicycle.'''
        # is there a data directory for this bicycle? if not, tell the user to
        # put some data in the folder so we have something to work with!
        try:
            pathToBicycle = os.path.join(pathToData, 'bicycles', bicycleName)
            if os.path.isdir(pathToBicycle):
                print("We have foundeth a directory named: " +
                      "{0}.".format(pathToBicycle))
                return super(Bicycle, cls).__new__(cls)
            else:
                raise ValueError
        except:
            mes = """Are you nuts?! Make a directory called '{0}' with basic
data for your bicycle in this directory: '{1}'. Then I can actually create a
bicycle object. You may either need to change to the correct directory or reset
the pathToData argument.""".format(bicycleName, pathToData)
            print(mes)
            return None

    def __init__(self, bicycleName, pathToData='.', forceRawCalc=False,
                 forcePeriodCalc=False):
        """
        Creates a bicycle object and sets the parameters based on the available
        data.

        Parameters
        ----------
        bicycleName : string
            The short name of your bicicleta. It should be one word with the
            first letter capitalized and all other letters lower case. You
            should have a matching directory under `<pathToData>/bicycles/`.
            For example: `<pathToData>/bicycles/Shortname`.
        pathToData : string
            This is the path to the folder where the bicycle/rider parameters
            and raw data are stored. The default is the current working
            directory.
        forceRawCalc : boolean
            Forces a recalculation of the benchmark parameters from the
            measured parameters. Otherwise it will only run the calculation if
            there is no benchmark parameter file.
        forcePeriodCalc : boolean
            Forces a recalculation of the periods from the oscillation data.

        Notes
        -----
        Bicycles are assumed not to have a rider when initially loaded.

        """

        self.bicycleName = bicycleName
        pathToBicycles = os.path.join(pathToData, 'bicycles')
        # the directory where the files for this bicycle are stored
        self.directory = os.path.join(pathToBicycles, bicycleName)

        # bicycles are assumed not to have a rider when initially loaded
        self.hasRider = False
        self.riderPar = {}
        self.human = None

        self.parameters = {}
        # if there are some parameter files, then load them
        if 'Parameters' in os.listdir(self.directory):
            parDir = os.path.join(self.directory, 'Parameters')
            parFiles = os.listdir(parDir)
            for parFile in parFiles:
                # remove the extension
                fname = os.path.splitext(parFile)[0]
                # get the bike and the parameter set type
                bike, ptype = io.space_out_camel_case(fname, output='list')
                # load the parameters
                pathToFile = os.path.join(parDir, parFile)
                self.parameters[ptype] = io.load_parameter_text_file(
                    pathToFile)

        # this is where the raw data files from the pendulum oscillations are
        # stored
        rawDataDir = os.path.join(self.directory, 'RawData')

        # it would be more robust to see if there are enough files in the
        # RawData directory, but that isn't implemented yet. For now you'll
        # just get and error sometime down the road when a period for the
        # missing files is needed.
        isRawDataDir = 'RawData' in os.listdir(self.directory)

        if isRawDataDir:
            print("Found the RawData directory:", rawDataDir)
            fname = bicycleName + 'Measured.txt'
            isMeasuredFile = fname in os.listdir(rawDataDir)
        else:
            isMeasuredFile = False

        isBenchmark = 'Benchmark' in self.parameters.keys()

        # the user wants to force a recalc and the data is there
        conOne = forceRawCalc and isRawDataDir and isMeasuredFile
        # the user doesn't want to force a recalc and there are no benchmark
        # parameters
        conTwo = not forceRawCalc and not isBenchmark

        if conOne or conTwo:
            print("Recalcuting the parameters.")
            par, extras = self.calculate_from_measured(
                    forcePeriodCalc=forcePeriodCalc)
            self.parameters['Benchmark'] = par
            self.extras = extras
            print("The glory of the %s parameters are upon you!"
                  % self.bicycleName)
        elif not forceRawCalc and isBenchmark:
            # we already have what we need
            stmt1 = "Looks like you've already got some parameters for %s, "
            stmt2 = "use forceRawCalc to recalculate."
            print((stmt1 + stmt2) % self.bicycleName)
            # load the measured.txt file if it exists
            pathToRawFile = os.path.join(rawDataDir, self.bicycleName +
                                         'Measured.txt')
            try:
                self.parameters['Measured'] = \
                        io.load_parameter_text_file(pathToRawFile)
            except IOError:
                pass
        else:
            print('''There is no data available. Create
            bicycles/{sn}/Parameters/{sn}Benchmark.txt and/or fill
            bicycle/{sn}/RawData/ with pendulum data mat files and the
            {sn}Measured.txt file'''.format(sn=bicycleName))

    def __str__(self):
        if self.hasRider:
            desc = "{0} with {1} on board.".format(self.bicycleName,
                                                   self.riderName)
        else:
            desc = "{0} with no one on board.".format(self.bicycleName)
        return desc

    def save_parameters(self, filetype='text'):
        """Saves all the parameter sets to file.

        Parameters
        ==========

        filetype : string, optional
            - 'text' : a text file with parameters as ``c = 0.10+/-0.01``
            - 'matlab' : matlab .mat file
            - 'pickle' : python pickled dictionary

        """
        if self.hasRider:
            pathToData = os.path.split(os.path.split(self.directory)[0])[0]
            pathToParDir = os.path.join(pathToData, 'riders', self.riderName,
                                        'Parameters')
            pathToCombDir = os.path.join(pathToParDir, 'Combined')
            if not os.path.exists(pathToCombDir):
                os.makedirs(pathToCombDir)
            fileName = self.riderName + self.bicycleName
            # don't resave the measured parameters
            psets = [x for x in self.riderPar.keys() if x != 'Measured']
            parameters = self.riderPar
            print(('This bicycle has a rider, {0}, so the data will be ' +
                   'saved here: {1}').format(self.riderName, pathToParDir))
        else:
            pathToParDir = os.path.join(self.directory, 'Parameters')
            fileName = self.bicycleName
            # don't resave the measured parameters
            psets = [x for x in self.parameters.keys() if x != 'Measured']
            parameters = self.parameters
            print(('This bicycle has no rider so the data will be ' +
                   'saved here: {0}').format(pathToParDir))

        if filetype == 'text':
            for pset in psets:
                fileName = fileName + pset + '.txt'
                pathToTxtFile = os.path.join(pathToParDir, fileName)
                io.write_parameter_text_file(pathToTxtFile, parameters[pset])
                if self.hasRider:
                    pathToCombFile = os.path.join(pathToCombDir, fileName)
                    io.write_parameter_text_file(pathToCombFile,
                                                 self.parameters[pset])

        elif filetype == 'matlab':
            # this should handle the uncertainties properly
            raise NotImplementedError("Doesn't work yet.")

        elif filetype == 'pickle':
            raise NotImplementedError("Doesn't work yet.")

    def show_pendulum_photos(self):
        """
        Opens up the pendulum photos in eye of gnome for inspection.

        This only works in Linux and if eog is installed. Maybe check pythons
        xdg-mime model for having this work cross platform.

        """
        photoDir = os.path.join(self.directory, 'Photos')
        try:
            if os.path.isdir(photoDir):
                os.system('eog ' + os.path.join(photoDir, '*.*'))
            else:
                print("There are no photos of your bicycle.")
        except:
            raise NotImplementedError("This works only works for linux with " +
                                      "Eye of Gnome installed.")

    def steer_assembly_moment_of_inertia(self, handlebar=True, fork=True,
                                         wheel=True, aboutSteerAxis=False,
                                         nominal=False):
        """
        Returns the inertia tensor of the steer assembly with respect to a
        reference frame aligned with the steer axis.

        Parameters
        ----------
        handlebar : boolean, optional
            If true the handlebar will be included in the calculation.
        fork : boolean, optional
            If true the fork will be included in the calculation.
        wheel : boolean, optional
            If true then the wheel will be included in the calculation.
        aboutSteerAxis : boolean, optional
            If true the inertia tensor will be with respect to a point made
            from the projection of the center of mass onto the steer axis.
        nominal : boolean, optional
            If true the nominal values will be returned instead of a uarray.

        Returns
        -------
        iAss : float
            Inertia tensor of the specified steer assembly parts with respect
            to a reference frame aligned with the steer axis.

        Notes
        -----
        The 3 component is aligned with the steer axis (pointing downward), the
        1 component is perpendicular to the steer axis (pointing forward) and
        the 2 component is perpendicular to the steer axis (pointing to the
        right).

        This function does not currently take into account the flywheel, D, if
        it is defined, beware.

        """
        # load in the Benchmark parameter set
        par = self.parameters['Benchmark']

        if 'mD' in par.keys():
            print("You have a flywheel defined. Beware that it is ignored in "
                  "the calculations and the results do not reflect that it is "
                  "there.")

        # there should always be either an H (handlebar/fork) and sometimes
        # there is a G (handlebar) and S (fork) if the fork and handlebar were
        # measured separately
        try:
            if fork and handlebar:
                # handlebar/fork
                I = inertia.part_inertia_tensor(par, 'H')
                m = par['mH']
                x = par['xH']
                z = par['zH']
            elif fork and not handlebar:
                # fork alone
                I = inertia.part_inertia_tensor(par, 'S')
                m = par['mS']
                x = par['xS']
                z = par['zS']
            elif handlebar and not fork:
                # handlebar alone
                I = inertia.part_inertia_tensor(par, 'G')
                m = par['mG']
                x = par['xG']
                z = par['zG']
            else:
                # if neither set to zero
                I = np.zeros((3, 3))
                m = 0.
                x = 0.
                z = 0.
        except KeyError:
            raise ValueError("The fork and handlebar were not measured " +
                             "separately for this bicycle." +
                             " Try making both the fork and handlebar either" +
                             " both True or both False.")

        if wheel:
            # list the mass and com of the handlebar/assembly and the front
            # wheel
            masses = np.array([m, par['mF']])

            coords = np.array([[x, par['w']],
                               [0., 0.],
                               [z, -par['rF']]])

            # mass and com of the entire assembly
            mAss, cAss = com.total_com(coords, masses)

            # front wheel inertia in the benchmark reference frame about the
            # com
            IF = inertia.part_inertia_tensor(par, 'F')

            # distance from the fork/handlebar assembly (without wheel) to the
            # new center of mass for the assembly with the wheel
            d = np.array([x - cAss[0], 0., z - cAss[2]])

            # distance from the front wheel center to the new center of mass
            # for the assembly with the wheel
            dF = np.array([par['w'] - cAss[0],
                           0.,
                           -par['rF'] - cAss[2]])

            # this is the inertia of the assembly about the com with reference
            # to the benchmark bicycle reference frame
            iAss = (inertia.parallel_axis(I, m, d) +
                    inertia.parallel_axis(IF, par['mF'], dF))

            # this is the inertia of the assembly about a reference frame
            # aligned with the steer axis and through the center of mass
            iAssRot = inertia.rotate_inertia_tensor(iAss, par['lam'])

        else:  # don't add the wheel
            mAss = m
            cAss = np.array([x, 0., z])
            iAssRot = inertia.rotate_inertia_tensor(I, par['lam'])

        if aboutSteerAxis:
            # this is the distance from the assembly com to the steer axis
            distance = geometry.distance_to_steer_axis(par['w'], par['c'],
                                                       par['lam'], cAss)
            print("handlebar cg distance", distance)

            # now calculate the inertia about the steer axis of the rotated
            # frame
            iAss = inertia.parallel_axis(iAssRot, mAss,
                                         np.array([distance, 0., 0.]))
        else:
            iAss = iAssRot

        if nominal:
            return unumpy.nominal_values(iAss)
        else:
            return iAss

    def calculate_from_measured(self, forcePeriodCalc=False):
        '''Calculates the parameters from measured data.'''

        rawDataDir = os.path.join(self.directory, 'RawData')
        pathToRawFile = os.path.join(rawDataDir,
                                     self.bicycleName + 'Measured.txt')

        # load the measured parameters
        self.parameters['Measured'] = io.load_parameter_text_file(
            pathToRawFile)

        forkIsSplit = is_fork_split(self.parameters['Measured'])

        # if the the user doesn't specifiy to force period calculation, then
        # see if enough data is actually available in the *Measured.txt file to
        # do the calculations
        if not forcePeriodCalc:
            forcePeriodCalc = period.check_for_period(
                self.parameters['Measured'], forkIsSplit)

        if forcePeriodCalc:
            # get the list of mat files associated with this bike
            matFiles = [x for x in os.listdir(rawDataDir)
                        if x.endswith('.mat')]
            matFiles.sort()
            # calculate the period for each file for this bicycle
            periods = period.calc_periods_for_files(rawDataDir, matFiles,
                                                    forkIsSplit)
            # add the periods to the measured parameters
            self.parameters['Measured'].update(periods)

            io.write_periods_to_file(pathToRawFile, periods)

        return calculate_benchmark_from_measured(self.parameters['Measured'])

    def add_rider(self, riderName, reCalc=False, draw=False):
        """
        Adds the inertial effects of a rigid rider to the bicycle.

        Parameters
        ----------
        riderName : string
            A rider name that corresponds to a folder in
            `<pathToData>/riders/`.
        reCalc : boolean, optional
            If true, the rider parameters will be recalculated.
        draw : boolean, optional
            If true, visual python will be used to draw a three dimensional
            image of the rider.

        """

        # can't draw the rider model without the human object
        if draw:
            reCalc = True

        # first check to see if a rider has already been added
        if self.hasRider:
            print(("D'oh! This bicycle already has {0} as a " +
                  "rider!").format(self.riderName))
        else:
            print("There is no rider on the bicycle, now adding " +
                  "{0}.".format(riderName))
            pathToData = os.path.split(os.path.split(self.directory)[0])[0]
            # get the path to the rider's folder
            pathToRider = os.path.join(pathToData, 'riders', riderName)
            # load in the parameters
            bicyclePar = self.parameters['Benchmark']
            bicycleName = self.bicycleName

            if reCalc:
                print("Calculating the human configuration.")
                # run the calculations
                try:
                    measuredPar = self.parameters['Measured']
                except KeyError:
                    print('The measured bicycle parameters need to be ' +
                          'available, create your bicycle such that they ' +
                          'are available.')
                    raise
                riderPar, human, bicycleRiderPar =\
                    rider.configure_rider(pathToRider, bicycleName, bicyclePar,
                                          measuredPar, draw)
            else:
                pathToParFile = os.path.join(pathToRider, 'Parameters',
                                             riderName + self.bicycleName +
                                             'Benchmark.txt')
                try:
                    # load the parameter file
                    riderPar = io.load_parameter_text_file(pathToParFile)
                except IOError:
                    # file doesn't exist so run the calculations
                    print("No parameter files found, calculating the human " +
                          "configuration.")
                    try:
                        measuredPar = self.parameters['Measured']
                    except KeyError:
                        print('The measured bicycle parameters need to be '
                              'available, create your bicycle such that they '
                              'are available.')
                        raise
                    riderPar, human, bicycleRiderPar =\
                        rider.configure_rider(pathToRider, bicycleName,
                                              bicyclePar, measuredPar, draw)
                else:
                    print("Loaded the precalculated parameters from " +
                          "{0}".format(pathToParFile))
                    bicycleRiderPar = inertia.combine_bike_rider(bicyclePar,
                                                                 riderPar)
            # set the attributes
            self.riderPar['Benchmark'] = riderPar
            try:
                self.human = human
            except NameError:
                self.human = None
            self.parameters['Benchmark'] = bicycleRiderPar
            self.riderName = riderName
            self.hasRider = True

    def plot_bicycle_geometry(self, show=True, pendulum=True,
                              centerOfMass=True, inertiaEllipse=True):
        """Returns a figure showing the basic bicycle geometry, the centers of
        mass and the moments of inertia.

        Parameters
        ==========
        show : boolean, optional
            If true ``matplotlib.pyplot.show()`` will be called before exiting
            the function.
        pendulum : boolean, optional
            If true the axes of the torsional pendulum will be displayed (only
            useful if raw measurement data is availabe).
        centerOfMass : boolean, optional
            If true the mass center of each rigid body will be displayed.
        inertiaEllipse : boolean optional
            If true inertia ellipses for each rigid body will be displayed.

        Returns
        =======
        fig : matplotlib.pyplot.Figure

        Notes
        =====
        If the flywheel is defined, it's center of mass corresponds to the
        front wheel and is not depicted in the plot.

        """
        par = io.remove_uncertainties(self.parameters['Benchmark'])
        parts = get_parts_in_parameters(par)

        try:
            slopes = io.remove_uncertainties(self.extras['slopes'])
            intercepts = io.remove_uncertainties(self.extras['intercepts'])
            penInertias = io.remove_uncertainties(
                self.extras['pendulumInertias'])
        except AttributeError:
            pendulum = False

        fig, ax = plt.subplots()

        fig.set_size_inches([4.0*GOLDEN_RATIO, 4.0])

        # define some colors for the parts
        numColors = len(parts)
        cmap = plt.get_cmap('gist_rainbow')
        partColors = {}
        for i, part in enumerate(parts):
            partColors[part] = cmap(1. * i / numColors)

        if inertiaEllipse:
            # plot the principal moments of inertia
            for j, part in enumerate(parts):
                I = inertia.part_inertia_tensor(par, part)
                Ip, C = inertia.principal_axes(I)
                if part == 'R':
                    center = np.array([0., par['rR']])
                elif part in 'FD':
                    center = np.array([par['w'], par['rF']])
                else:
                    center = np.array([par['x' + part], -par['z' + part]])
                # which row in C is the y vector
                uy = np.array([0., 1., 0.])
                for i, row in enumerate(C):
                    if np.abs(np.sum(row - uy)) < 1E-10:
                        yrow = i
                # remove the row for the y vector
                Ip2D = np.delete(Ip, yrow, 0)
                # remove the column and row associated with the y
                C2D = np.delete(np.delete(C, yrow, 0), 1, 1)
                # make an ellipse
                Imin = Ip2D[0]
                Imax = Ip2D[1]
                # get width and height of a ellipse with the major axis equal
                # to one
                unitWidth = 1. / 2. / np.sqrt(Imin) * np.sqrt(Imin)
                unitHeight = 1. / 2. / np.sqrt(Imax) * np.sqrt(Imin)
                # now scaled the width and height relative to the maximum
                # principal moment of inertia
                width = Imax * unitWidth
                height = Imax * unitHeight
                angle = -np.degrees(np.arccos(C2D[0, 0]))
                ellipse = Ellipse((center[0], center[1]), width, height,
                                  angle=angle, fill=False,
                                  color=partColors[part], alpha=0.25)
                ax.add_patch(ellipse)

        # plot the ground line
        x = np.array([-par['rR'],
                      par['w'] + par['rF']])
        ax.plot(x, np.zeros_like(x), 'k')

        # plot the rear wheel
        c = plt.Circle((0., par['rR']), radius=par['rR'], fill=False)
        ax.add_patch(c)

        # plot the front wheel
        c = plt.Circle((par['w'], par['rF']), radius=par['rF'], fill=False)
        ax.add_patch(c)

        # plot the fundamental bike
        deex, deez = geometry.fundamental_geometry_plot_data(par)
        ax.plot(deex, -deez, 'k')

        # plot the steer axis
        dx3 = deex[2] + deez[2] * (deex[2] - deex[1]) / (deez[1] - deez[2])
        ax.plot([deex[2], dx3],  [-deez[2], 0.], 'k--')

        # don't plot the pendulum lines if a rider has been added because the
        # inertia has changed
        if self.hasRider:
            pendulum = False

        if pendulum:
            # plot the pendulum axes for the measured parts
            for j, pair in enumerate(slopes.items()):
                part, slopeSet = pair
                xcom, zcom = par['x' + part], par['z' + part]
                for i, m in enumerate(slopeSet):
                    b = intercepts[part][i]
                    xPoint, zPoint = geometry.project_point_on_line((m, b),
                                                                    (xcom,
                                                                     zcom))
                    comLineLength = penInertias[part][i]
                    xPlus = comLineLength / 2. * np.cos(np.arctan(m))
                    x = np.array([xPoint - xPlus,
                                  xPoint + xPlus])
                    z = -m * x - b
                    ax.plot(x, z, color=partColors[part])
                    # label the pendulum lines with a number
                    ax.text(x[0], z[0], str(i + 1))

        if centerOfMass:
            # plot the center of mass location
            def com_symbol(ax, center, radius, color='b'):
                '''Returns axis with center of mass symbol.'''
                c = plt.Circle(center, radius=radius, fill=False)
                w1 = Wedge(center, radius, 0., 90.,
                           color=color, ec=None, alpha=0.5)
                w2 = Wedge(center, radius, 180., 270.,
                           color=color, ec=None, alpha=0.5)
                ax.add_patch(w1)
                ax.add_patch(w2)
                ax.add_patch(c)
                return ax

            # radius of the CoM symbol
            sRad = 0.03
            # front wheel CoM
            ax = com_symbol(ax, (par['w'], par['rF']), sRad,
                            color=partColors['F'])
            ax.text(par['w'] + sRad, par['rF'] + sRad, 'F')
            # rear wheel CoM
            ax = com_symbol(ax, (0., par['rR']), sRad,
                            color=partColors['R'])
            ax.text(0. + sRad, par['rR'] + sRad, 'R')
            for j, part in enumerate([x for x in parts
                                      if x not in 'RFD']):
                xcom = par['x' + part]
                zcom = par['z' + part]
                ax = com_symbol(ax, (xcom, -zcom), sRad,
                                color=partColors[part])
                ax.text(xcom + sRad, -zcom + sRad, part)
            if 'H' not in parts:
                ax = com_symbol(ax, (par['xH'], -par['zH']), sRad)
                ax.text(par['xH'] + sRad, -par['zH'] + sRad, 'H')

        # if there is a rider on the bike, make a simple stick figure
        top_of_head = 0.0
        if self.human:
            human = self.human
            mpar = self.parameters['Measured']
            bpar = self.parameters['Benchmark']
            # K2: lower leg, tip of foot to knee
            start = rider.yeadon_vec_to_bicycle_vec(human.K2.end_pos, mpar,
                                                    bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.K2.pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            # K1: upper leg, knee to hip
            start = rider.yeadon_vec_to_bicycle_vec(human.K2.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.K1.pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            # torso
            start = rider.yeadon_vec_to_bicycle_vec(human.K1.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.B1.pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            # B1: upper arm
            start = rider.yeadon_vec_to_bicycle_vec(human.B1.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.B2.pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            # B2: lower arm, elbow to tip of fingers
            start = rider.yeadon_vec_to_bicycle_vec(human.B2.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.B2.end_pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            # C: chest/head
            start = rider.yeadon_vec_to_bicycle_vec(human.B1.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.C.end_pos, mpar, bpar)
            ax.plot([start[0, 0], end[0, 0]],
                    [-start[2, 0], -end[2, 0]], 'k')
            top_of_head = -end[2, 0]

        ax.set_aspect('equal')

        # set the y limits to encompass the bicycle and rider geometry
        max_y = max([2*par['rR'],  # rear wheel diameter
                     2*par['rF'],  # front wheel diameter
                     max(-deez),  # max of Z values of bicycle geometry
                     top_of_head])  # max of Z values of human head
        min_y = min(-deez)
        if min_y >= 0.0:
            y_low = min([0.0, min_y])
        else:
            y_low = -np.ceil(np.abs(min_y))
        ax.set_ylim((y_low, np.ceil(max_y)))

        ax.set_title("{}\nBicycle Geometry".format(self.bicycleName))

        ax.set_xlabel('x [m]')
        ax.set_ylabel('-z [m]')

        if show:
            fig.show()

        # TODO : This should return ax instead of fig to follow typical
        # practice in other Python libraries.

        return fig

    def _plot_bicycle_geometry_plotly(self, show=True, pendulum=True,
                                      centerOfMass=True, inertiaEllipse=True):
        """Returns a Plotly figure showing the basic bicycle geometry,
        the centers of
        mass and the moments of inertia.

        Parameters
        ==========
        show : optional
            If true plotly figure will show.
        centerOfMass : boolean, optional
            If true the mass center of each rigid body will be displayed. but
            will have no trace in the legend since it already has a button.
            The hoverfunction will show where the COM's are located.
        inertiaEllipse : boolean optional
            If true inertia ellipses for each rigid body will be displayed.
            In some cases the ellipses are so large that they will not fit in
            the figure. Therefor the axis of this plot are fixed.

        Returns
        =======
        fig1 : A plotly figure

        """
        if px is None:
            raise ImportError('plotly is not installed')
        par = io.remove_uncertainties(self.parameters['Benchmark'])
        parts = get_parts_in_parameters(par)

        try:
            slopes = io.remove_uncertainties(self.extras['slopes'])
            intercepts = io.remove_uncertainties(self.extras['intercepts'])
            penInertias = io.remove_uncertainties(
                self.extras['pendulumInertias'])
        except AttributeError:
            pendulum = False

        fig1 = go.Figure()

        # define some colors for the parts
        # cmap = px.colors.sequential.Agsunset
        cmap = px.colors.qualitative.Pastel
        partColors = {}

        for i, part in enumerate(parts):
            partColors[part] = cmap[i]

        if inertiaEllipse:
            # plot the principal moments of inertia
            for j, part in enumerate(parts):
                I = inertia.part_inertia_tensor(par, part)
                Ip, C = inertia.principal_axes(I)
                if part == 'R':
                    center = np.array([0., par['rR']])
                elif part in 'FD':
                    center = np.array([par['w'], par['rF']])
                else:
                    center = np.array([par['x' + part], -par['z' + part]])
                    # which row in C is the y vector
                uy = np.array([0., 1., 0.])
                for i, row in enumerate(C):
                    if np.abs(np.sum(row - uy)) < 1E-10:
                        yrow = i
                    # remove the row for the y vector
                Ip2D = np.delete(Ip, yrow, 0)
                # remove the column and row associated with the y
                C2D = np.delete(np.delete(C, yrow, 0), 1, 1)
                # make an ellipse
                Imin = Ip2D[0]
                Imax = Ip2D[1]
                # get width and height of a ellipse with
                # the major axis equal to one
                unitWidth = 1. / 2. / np.sqrt(Imin) * np.sqrt(Imin)
                unitHeight = 1. / 2. / np.sqrt(Imax) * np.sqrt(Imin)
                # now scaled the width and height relative to
                # the maximum principal moment of inertia
                width = Imax * unitWidth
                height = Imax * unitHeight
                angle = -np.degrees(np.arccos(C2D[0, 0]))
                x_center = center[0]
                y_center = center[1]
                x_ep, y_ep = plot._generate_ellipse_plot_data(
                    x_center=x_center, y_center=y_center,
                    ax1=[np.cos(angle), np.sin(angle)],
                    ax2=[-np.sin(angle), np.cos(angle)],
                    a=height, b=width, N=100)

                fig1.add_scatter(x=x_ep, y=y_ep, mode='lines',
                                 name='Inertia of ' + part,
                                 line_color=partColors[part],
                                 fill='toself', opacity=0.5)

        # plot the ground line
        x = np.array([-par['rR'], par['w'] + par['rF']])
        fig1.add_trace(go.Scatter(x=x, y=np.zeros_like(x),
                       mode='lines',
                       name='Ground',
                       line_color='lightgrey',
                       hovertemplate="%{x:.3f}<br>%{y:.3f}"))

        def make_circle_legend(R, x_center_wheel, y_center_wheel):
            t = np.linspace(0, 2*np.pi, 100)
            xwh = R*np.cos(t)
            ywh = R*np.sin(t)
            x_wheel = xwh + x_center_wheel
            y_wheel = ywh + y_center_wheel
            return x_wheel, y_wheel

        # plot the rear wheel
        x_wheel_R, y_wheel_R = make_circle_legend(par['rR'], 0, par['rR'])
        fig1.add_trace(go.Scatter(x=x_wheel_R,
                                  y=y_wheel_R,
                                  mode='lines',
                                  line_color='grey',
                                  name='Rear wheel',
                                  hovertemplate="%{x:.3f}<br>%{y:.3f}"))

        # plot the front wheel
        x_wheel_F, y_wheel_F = make_circle_legend(par['rF'], par['w'],
                                                  par['rF'])
        fig1.add_trace(go.Scatter(x=x_wheel_F,
                                  y=y_wheel_F,
                                  mode='lines',
                                  line_color='grey',
                                  hovertemplate="%{x:.3f}<br>%{y:.3f}",
                                  name='Front wheel'))

        # plot the fundamental bike
        deex, deez = geometry.fundamental_geometry_plot_data(par)
        fig1.add_trace(go.Scatter(x=deex, y=-deez,
                                  mode='lines',
                                  name='Bicycle',
                                  line_color='black',
                                  hovertemplate="%{x:.3f}<br>%{y:.3f}"))

        # plot the steer axis
        dx3 = deex[2] + deez[2] * (deex[2] - deex[1]) / (deez[1] - deez[2])

        fig1.add_trace(go.Scatter(x=[deex[2], dx3], y=[-deez[2], 0.],
                                  mode='lines',
                                  name='Steer axis',
                                  hovertemplate="%{x:.3f}<br>%{y:.3f}",
                                  line=dict(dash='dash', color='dodgerblue')))

        # Update Layout so circle will be round and background white and no
        # grid
        fig1.update_xaxes(showgrid=False, zeroline=False)
        fig1.update_yaxes(showgrid=False, zeroline=False)

        fig1.update_xaxes(
            range=[-par['rR'], par['w']+par['rF']],  # sets the range of xaxis
            constrain="domain",  # meanwhile compresses the xaxis by decreasing its "domain"
        )
        fig1.update_yaxes(scaleanchor="x", scaleratio=1)
        fig1.update_layout(plot_bgcolor="white")

        # don't plot the pendulum lines if a rider has been added because the
        # inertia has changed
        if self.hasRider:
            pendulum = False

        if pendulum:
            # plot the pendulum axes for the measured parts
            for j, pair in enumerate(slopes.items()):
                part, slopeSet = pair
                xcom, zcom = par['x' + part], par['z' + part]
                for i, m in enumerate(slopeSet):
                    b = intercepts[part][i]
                    xPoint, zPoint = geometry.project_point_on_line((m, b),
                                                                    (xcom,
                                                                     zcom))
                    comLineLength = penInertias[part][i]
                    xPlus = comLineLength / 2. * np.cos(np.arctan(m))
                    xp = np.array([xPoint - xPlus,
                                  xPoint + xPlus])
                    zp = -m*xp - b

                    fig1.add_scatter(x=xp, y=zp, mode='lines', name='Pendulum',
                                     line=dict(dash='dash'))

        if centerOfMass:
            def com_symbol(R, x_center, y_center, partcolor):
                t = np.linspace(0, 0.5*np.pi, 100)
                xs = R*np.cos(t)
                ys = R*np.sin(t)
                xc1 = xs + x_center
                yc1 = ys + y_center
                t2 = np.linspace(0.5*np.pi, np.pi, 100)
                xs2 = R*np.cos(t2)
                ys2 = R*np.sin(t2)
                xc2 = xs2 + x_center
                yc2 = ys2 + y_center
                t3 = np.linspace(np.pi, 1.5*np.pi, 100)
                xs3 = R*np.cos(t3)
                ys3 = R*np.sin(t3)
                xc3 = xs3 + x_center
                yc3 = ys3 + y_center
                t4 = np.linspace(1.5*np.pi, 2*np.pi, 100)
                xs4 = R*np.cos(t4)
                ys4 = R*np.sin(t4)
                xc4 = xs4 + x_center
                yc4 = ys4 + y_center

                fig1.add_trace(go.Scatter(x=[x_center, x_center+R],
                                          y=[y_center, y_center], mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=[x_center, x_center],
                                          y=[y_center, y_center + R],
                                          mode='lines', line_color=partcolor,
                                          showlegend=False,
                                          hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=xc1, y=yc1, mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, fill='tonexty',
                                          hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=xc2, y=yc2, mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=xc3, y=yc3, mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=[x_center-R, x_center],
                                          y=[y_center, y_center], mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, fill='tonexty',
                                          hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=[x_center, x_center],
                                          y=[y_center-R, y_center],
                                          mode='lines', line_color=partcolor,
                                          showlegend=False,
                                          hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=xc4, y=yc4, mode='lines',
                                          line_color=partcolor,
                                          showlegend=False, hoverinfo='none'))
                fig1.add_trace(go.Scatter(x=[x_center, x_center],
                                          y=[y_center, y_center],
                                          mode='lines', line_color=partcolor,
                                          hovertemplate="%{x:.3f}<br>%{y:.3f}",
                                          name='COM', showlegend=False))
                return fig1

            # radius of the CoM symbol
            sRad = 0.03
            # front wheel CoM
            x_com_Wf = par['w']
            y_com_Wf = par['rF']
            fig1 = com_symbol(sRad, x_com_Wf, y_com_Wf, partColors['F'])
            fig1.add_annotation(text='F', xref='x', yref='y',
                                x=x_com_Wf + 0.055,
                                y=y_com_Wf+0.055, showarrow=False,
                                font=dict(size=15))

            # rear wheel CoM
            fig1 = com_symbol(sRad, 0., par['rR'], partColors['R'])
            fig1.add_annotation(text='R', xref='x', yref='y', x=0.055,
                                y=par['rR']+0.055, showarrow=False,
                                font=dict(size=15))

            for j, part in enumerate([x for x in parts
                                      if x not in 'RFD']):
                xcom = par['x' + part]
                zcom = par['z' + part]
                fig1 = com_symbol(sRad, xcom, -zcom, partColors[part])
                fig1.add_annotation(text=part, xref='x', yref='y',
                                    x=xcom+0.055, y=-zcom+0.055,
                                    showarrow=False, font=dict(size=15))

            if 'H' not in parts:
                fig1 = com_symbol(sRad, par['xH'], -par['zH'], partColors['H'])
                fig1.add_annotation(text="H", xref='x', yref='y',
                                    x=par['xH']+0.055, y=-par['zH']+0.055,
                                    showarrow=False, font=dict(size=15))

        # if there is a rider on the bike, make a simple stick figure
        if self.human:
            human = self.human
            mpar = self.parameters['Measured']
            bpar = self.parameters['Benchmark']
            # K2: lower leg, tip of foot to knee
            start = rider.yeadon_vec_to_bicycle_vec(human.K2.end_pos, mpar,
                                                    bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.K2.pos, mpar, bpar)
            fig1.add_trace(go.Scatter(x=[start[0, 0], end[0, 0]],
                                      y=[-start[2, 0], -end[2, 0]],
                                      mode='lines'))

            # K1: upper leg, knee to hip
            start = rider.yeadon_vec_to_bicycle_vec(human.K2.pos, mpar, bpar)
            end = rider.yeadon_vec_to_bicycle_vec(human.K1.pos, mpar, bpar)

        fig1.update_layout(title=dict(text='Bicycle geometry',
                           font=dict(family="Segoe UI", size=25)),
                           yaxis_title='z-axis [m]',
                           font_family="Source Sans Pro",
                           hoverlabel=dict(font_family="Source Sans Pro"))

        fig1.update_layout(yaxis=dict(autorange=True, showgrid=False,
                                      ticks='outside', showticklabels=True))
        fig1.update_layout(xaxis=dict(autorange=True, showgrid=False, ticks='',
                                      showticklabels=False))
        if show:
            fig1.show()
        return fig1

    def canonical(self, nominal=False):
        """
        Returns the canonical velocity and gravity independent matrices for
        the Whipple bicycle model linearized about the nominal
        configuration.

        Parameters
        ----------
        nominal : boolean, optional
            The default is false and uarrays are returned with the
            calculated uncertainties. If true ndarrays are returned without
            uncertainties.

        Returns
        -------
        M : uarray, shape(2,2)
            Mass matrix.
        C1 : uarray, shape(2,2)
            Velocity independent damping matrix.
        K0 : uarray, shape(2,2)
            Gravity independent part of the stiffness matrix.
        K2 : uarray, shape(2,2)
            Velocity squared independent part of the stiffness matrix.

        Notes
        -----

        The canonical matrices complete the following equation:

            M * q'' + v * C1 * q' + [g * K0 + v**2 * K2] * q = f

        where:

            q = [phi, delta]
            f = [Tphi, Tdelta]

        phi
            Bicycle roll angle.
        delta
            Steer angle.
        Tphi
            Roll torque.
        Tdelta
            Steer torque.
        v
            Bicylce speed.

        If you have a flywheel defined, body D, it will completely be
        ignored in these results. These results are strictly for the Whipple
        bicycle model.

        """

        par = self.parameters['Benchmark']

        M, C1, K0, K2 = bicycle.benchmark_par_to_canonical(par)

        if nominal is True:
            return (unumpy.nominal_values(M),
                    unumpy.nominal_values(C1),
                    unumpy.nominal_values(K0),
                    unumpy.nominal_values(K2))
        elif nominal is False:
            return M, C1, K0, K2
        else:
            raise ValueError('nominal must be True or False')

    def state_space(self, speed, nominal=False):
        """
        Returns the A and B matrices for the Whipple model linearized about
        the upright constant velocity configuration.


        Parameters
        ----------
        speed : float
            The speed of the bicycle.
        nominal : boolean, optional
            The default is false and uarrays are returned with the calculated
            uncertainties. If true ndarrays are returned without uncertainties.

        Returns
        -------

        A : ndarray, shape(4,4)
            The state matrix.
        B : ndarray, shape(4,2)
            The input matrix.

        Notes
        -----
        ``A`` and ``B`` describe the Whipple model in state space form:

            x' = A * x + B * u

        where

        The states are [roll angle,
                        steer angle,
                        roll rate,
                        steer rate]

        The inputs are [roll torque,
                        steer torque]

        If you have a flywheel defined, body D, it will completely be ignored
        in these results. These results are strictly for the Whipple bicycle
        model.

        """

        M, C1, K0, K2 = self.canonical()

        g = self.parameters['Benchmark']['g']

        A, B = bicycle.ab_matrix(M, C1, K0, K2, speed, g)

        if nominal is True:
            return (unumpy.nominal_values(A), unumpy.nominal_values(B))
        elif nominal is False:
            return A, B
        else:
            raise ValueError('nominal must be True or False')

    def eig(self, speeds):
        '''Returns the eigenvalues and eigenvectors of the Whipple bicycle
        model linearized about the nominal configuration.

        Parameters
        ----------
        speeds : ndarray, shape (n,) or float
            The speed at which to calculate the eigenvalues.

        Returns
        -------
        evals : ndarray, shape (n, 4)
            eigenvalues
        evecs : ndarray, shape (n, 4, 4)
            eigenvectors

        Notes
        -----
        If you have a flywheel defined, body D, it will completely be ignored
        in these results. These results are strictly for the Whipple bicycle
        model.

        '''
        # this allows you to enter a float
        try:
            speeds.shape
        except AttributeError:
            speeds = np.array([speeds])

        par = io.remove_uncertainties(self.parameters['Benchmark'])

        M, C1, K0, K2 = bicycle.benchmark_par_to_canonical(par)

        m, n = 4, speeds.shape[0]
        evals = np.zeros((n, m), dtype='complex128')
        evecs = np.zeros((n, m, m), dtype='complex128')
        for i, speed in enumerate(speeds):
            A, B = bicycle.ab_matrix(M, C1, K0, K2, speed, par['g'])
            w, v = np.linalg.eig(A)
            evals[i] = w
            evecs[i] = v

        return evals, evecs

    def plot_eigenvalues_vs_speed(self, speeds, fig=None, generic=False,
                                  color='black', show=False, largest=False,
                                  linestyle='-', grid=False, show_legend=True):
        """Returns a plot of the eigenvalues versus speed for the current
        benchmark parameters.

        Parameters
        ----------
        speeds : ndarray, shape(n,)
            An array of speeds to calculate the eigenvalues at.
        fig : matplotlib figure, optional
            A figure to plot to.
        generic : boolean
            If true the lines will all be the same color and the modes will
            not be labeled.
        color : matplotlib color
            If generic is true this will be the color of the plot lines.
        largest : boolean
            If true, only the largest eigenvalue is plotted.
        grid : boolean, optional
            If true, displays a grid on the plot.
        show_legend: boolean, optional
            If true, displays a legend describing the different parts of the
            solution shown.

        Returns
        -------
        fig : matpolib.pyplot.Figure
            The figure.

        Notes
        -----
        If you have a flywheel defined, body D, it will completely be
        ignored in these results. These results are strictly for the Whipple
        bicycle model.

        """

        # sort the speeds in case they aren't
        speeds = np.sort(speeds)

        # figure properties
        fig_height = 4.0  # inches
        figsize = [fig_height*GOLDEN_RATIO, fig_height]
        params = {
            'axes.labelsize': 8,
            'text.fontsize': 10,
            'legend.fontsize': 8,
            'xtick.labelsize': 6,
            'ytick.labelsize': 6,
            'figure.figsize': figsize
            }
        # NOTE : text.fontsize no longer supported in matplotlib
        try:
            plt.rcParams.update(params)
        except KeyError:
            del params['text.fontsize']
            params['font.size'] = 10
            plt.rcParams.update(params)

        if fig is None:
            fig, ax = plt.subplots(figsize=figsize)

        evals, evecs = self.eig(speeds)

        if largest:
            generic = True

        if generic:
            weaveColor = color
            capsizeColor = color
            casterColor = color
            legend = ['_nolegend_'] * 6
            legend[5] = self.bicycleName
            maxLabel = self.bicycleName
        else:
            weaveColor = 'blue'
            capsizeColor = 'red'
            casterColor = 'green'
            legend = ['Imaginary Weave', 'Imaginary Capsize',
                      'Imaginary Caster', 'Real Weave', 'Real Capsize',
                      'Real Caster']
            maxLabel = 'Max Eigenvalue'

        if largest:
            maxEval = np.max(np.real(evals), axis=1)
            ax.plot(speeds, maxEval, color=color, label=maxLabel,
                    linestyle=linestyle, linewidth=1.5)
            # x axis line
            ax.plot(speeds, np.zeros_like(speeds), 'k-', label='_nolegend_',
                    linewidth=1.5)
            ax.set_ylim((np.min(maxEval), np.max(maxEval)))
            ax.set_ylabel('Real Part of the Largest Eigenvalue [1/s]')
        else:
            wea, cap, cas = bicycle.sort_modes(evals, evecs)

            # imaginary components
            ax.plot(speeds, np.abs(np.imag(wea['evals'])), color=weaveColor,
                    label=legend[0], linestyle='--')
            ax.plot(speeds, np.abs(np.imag(cap['evals'])), color=capsizeColor,
                    label=legend[1], linestyle='--')
            ax.plot(speeds, np.abs(np.imag(cas['evals'])), color=casterColor,
                    label=legend[2], linestyle='--')

            # x axis line
            ax.plot(speeds, np.zeros_like(speeds), 'k-', label='_nolegend_',
                    linewidth=1.5)

            # plot the real parts of the eigenvalues
            ax.plot(speeds, np.real(wea['evals']), color=weaveColor,
                    label=legend[3])
            ax.plot(speeds, np.real(cap['evals']), color=capsizeColor,
                    label=legend[4])
            ax.plot(speeds, np.real(cas['evals']), color=casterColor,
                    label=legend[5])

            # set labels and limits
            ax.set_ylabel('Real and Imaginary Parts of the Eigenvalue [1/s]')

        ax.set_xlim((speeds[0], speeds[-1]))
        ax.set_xlabel('Speed [m/s]')

        if generic:
            ax.set_title('Eigenvalues vs Speed')
        else:
            ax.set_title('%s\nEigenvalues vs Speed' % self.bicycleName)
            if show_legend:
                ax.legend()

        if grid:
            ax.grid()

        if show:
            fig.show()

        return fig

    def _plot_eigenvalues_vs_speed_plotly(self, speeds, fig=None, show=True,
                                          largest=False,
                                          stability_region=True):
        if px is None:
            raise ImportError('plotly is not installed')
        speeds = np.sort(speeds)
        if fig is None:
            fig = go.Figure(layout_yaxis_range=[-10, 10])
            evals, evecs = self.eig(speeds)

        if largest:
            fig.add_trace(go.Scatter(x=speeds, y=np.max(evals)))
            fig.show()
        else:
            w, cap, cas = bicycle.sort_modes(evals, evecs)
            colors_eig = px.colors.qualitative.Pastel
            weaveColor1 = colors_eig[0]
            weaveColor2 = colors_eig[1]
            capsizeColor = colors_eig[2]
            casterColor = colors_eig[5]
        wea1 = w['evals'][:, 0]
        wea2 = w['evals'][:, 1]
        fig.add_trace(go.Scatter(x=speeds, y=np.real(wea1),
                                 mode='lines',
                                 name='Re Weave',
                                 line=dict(color=weaveColor1),
                                 text='Re'))
        fig.add_trace(go.Scatter(x=speeds, y=np.real(wea2),
                                 mode='lines',
                                 name='Re Weave',
                                 line=dict(color=weaveColor2),
                                 text='Re'))
        fig.add_trace(go.Scatter(x=speeds, y=np.real(cap['evals']),
                                 mode='lines',
                                 name='Re Capsize',
                                 line=dict(color=capsizeColor),
                                 text='Re'))
        fig.add_trace(go.Scatter(x=speeds, y=np.real(cas['evals']),
                                 mode='lines',
                                 name='Re Castering',
                                 line=dict(color=casterColor),
                                 text='Re'))
        fig.add_trace(go.Scatter(x=speeds, y=np.abs(np.imag(wea1)),
                                 mode='lines',
                                 name='Im Weave',
                                 line=dict(color=weaveColor1, dash='dash'),
                                 text='Im'))
        fig.add_trace(go.Scatter(x=speeds, y=np.abs(np.imag(wea2)),
                                 mode='lines',
                                 name='Im Weave',
                                 line=dict(color=weaveColor2, dash='dash'),
                                 text='Im'))
        fig.add_trace(go.Scatter(x=speeds, y=np.abs(np.imag(cap['evals'])),
                                 mode='lines',
                                 name='Im Capsize',
                                 line=dict(color=capsizeColor, dash='dash'),
                                 text='Im'))
        fig.add_trace(go.Scatter(x=speeds, y=np.abs(np.imag(cas['evals'])),
                                 mode='lines',
                                 name='Im Castering',
                                 line=dict(color=casterColor, dash='dash'),
                                 text='Im'))
        if stability_region:
            try:
                v_start_stab = max([min(speeds[np.real(wea2) < 0]),
                                    min(speeds[np.real(cas['evals']) < 0]),
                                    min(speeds[np.real(cap['evals']) < 0]),
                                    min(speeds[np.real(wea1) < 0],
                                        default="EMPTY")])
                v_end_stab = min([max(speeds[np.real(wea2) < 0]),
                                  max(speeds[np.real(cas['evals']) < 0]),
                                  max(speeds[np.real(cap['evals']) < 0]),
                                  max(speeds[np.real(wea1) < 0])])
            except:  # TODO : Add explicit exception
                fig.add_annotation(x=0.5*max(speeds), y=9,
                                   text="No stability region",
                                   showarrow=False)
            if (v_start_stab > v_end_stab):
                fig.add_annotation(x=0.5*max(speeds), y=9,
                                   text="No stability region",
                                   showarrow=False)
            elif (v_end_stab - v_start_stab < 0.0001):
                fig.add_annotation(x=0.5*max(speeds), y=9,
                                   text="No stability region",
                                   showarrow=False)
            else:
                fig.add_vrect(x0=v_start_stab, x1=v_end_stab,
                              annotation_text="Self stability",
                              annotation_position='top left',
                              # fillcolor="blue", opacity=0.25,
                              fillcolor='rgba(71,147,231,0.5)',
                              line_width=0, row=1, col=1)

        fig.update_layout(title=dict(text='Eigenvalues vs velocity',
                          font=dict(family="Segoe UI", size=25)),
                          font_family="Source Sans Pro",
                          plot_bgcolor='rgba(39,128,227,0.15)',
                          xaxis_title='Velocity [m/s]',
                          yaxis_title='Eigenvalues [1/s]',
                          hoverlabel=dict(font_family="Source Sans Pro"))
        fig.update_traces(hovertemplate="%{x:.3f}<br>%{y:.3f}")

        if show:
            fig.show()

        return fig

    def plot_bode(self, speed, u, y, **kwargs):
        """Returns a Bode plot.

        Parameters
        ----------
        speed : float
            The speed at which to evaluate the system.
        u : integer
            An integer between 0 and 1 corresponding to the inputs roll torque
            and steer torque.
        y : integer
            An integer between 0 and 3 corresponding to the inputs roll angle
            steer angle, roll rate, steer rate.
        kwargs : keyword pairs
            Any options that can be passed to dtk.bode.

        Returns
        -------
        mag : ndarray, shape(1000,)
            The magnitude in dB of the frequency reponse.
        phase : ndarray, shape(1000,)
            The phase in degress of the frequency response.
        fig : matplotlib figure
            The Bode plot.

        """

        A, B = self.state_space(speed, nominal=True)

        C = np.eye(A.shape[0])
        D = np.zeros_like(B)

        w = np.logspace(0, 2, 1000)

        outputNames = ['Roll Angle', 'Steer Angle', 'Roll Rate', 'Steer Rate']
        inputNames = ['Roll Torque', 'Steer Torque']

        if 'title' not in kwargs.keys():
            kwargs['title'] = inputNames[u] + ' to ' + outputNames[y]

        bode = control.bode((A, B[:, u], C[y, :], D[y, u]), w, **kwargs)

        return bode

    def compare_bode_speeds(self, speeds, u, y, fig=None):
        """Returns a figure with the Bode plots of multiple bicycles.

        Parameters
        ----------
        speeds : list
            A list of speeds at which to evaluate the system.
        u : integer
            An integer between 0 and 1 corresponding to the inputs roll torque
            and steer torque.
        y : integer
            An integer between 0 and 3 corresponding to the inputs roll angle,
            steer angle, roll rate, steer rate.

        Returns
        -------
        fig : matplotlib.Figure instance
            The Bode plot.

        Notes
        -----
        The phases are matched around zero degrees at with respect to the first
        frequency.

        """

        if fig is None:
            fig = plt.figure()

        for speed in speeds:
            self.plot_bode(speed, u, y, label=str(speed) + ' m/s', fig=fig)

        # take care of phase misalignment
        phaseLines = fig.ax2.lines
        for line in phaseLines:
            firstValue = line.get_ydata()[0]
            n = np.ceil(np.floor(abs(firstValue / 180.)) / 2.)
            line.set_ydata(line.get_ydata() - np.sign(firstValue) * n * 360.)
        return fig


def get_parts_in_parameters(par):
    '''Returns a list of parts in a parameter dictionary.

    Parameters
    ----------
    par : dictionary
        Benchmark bicycle parameters.

    Returns
    -------
    parts : list
        Unique list of parts that contain one or more of 'H', 'B', 'F', 'R',
        'S', 'G', 'D'.

    '''
    parts = [x[1] for x in par.keys() if x.startswith('m')]
    return parts


def calculate_benchmark_from_measured(mp):
    '''Returns the benchmark (Meijaard 2007) parameter set based on the
    measured data.

    Parameters
    ----------
    mp : dictionary
        Complete set of measured data.

    Returns
    -------
    par : dictionary
        Benchmark bicycle parameter set.

    '''

    forkIsSplit = is_fork_split(mp)

    par = {}

    # calculate the wheelbase, steer axis tilt and trail
    par = geometry.calculate_benchmark_geometry(mp, par)

    # masses
    par['mB'] = mp['mB']
    par['mF'] = mp['mF']
    par['mR'] = mp['mR']
    try:
        # we measured the mass of the flywheel plus the mass of the front
        # wheel, mp['mD'], so to get the actual mass of the flywheel, subtract
        # the mass of the front wheel
        par['mD'] = mp['mD'] - mp['mF']
    except KeyError:
        pass
    if forkIsSplit:
        par['mS'] = mp['mS']
        par['mG'] = mp['mG']
    else:
        par['mH'] = mp['mH']

    # get the slopes, intercepts and betas for each part
    slopes, intercepts, betas = com.part_com_lines(mp, par, forkIsSplit)

    # calculate the centers of mass
    for part in slopes.keys():
        par['x' + part], par['z' + part] = com.center_of_mass(slopes[part],
                                                              intercepts[part])

    # find the center of mass of the handlebar/fork assembly if the fork was
    # split
    if forkIsSplit:
        coordinates = np.array([[par['xS'], par['xG']],
                                [0., 0.],
                                [par['zS'], par['zG']]])
        masses = np.array([par['mS'], par['mG']])
        mH, cH = inertia.total_com(coordinates, masses)
        par['mH'] = mH
        par['xH'] = cH[0]
        par['zH'] = cH[2]

    # local accelation due to gravity
    par['g'] = mp['g']

    # calculate the wheel y inertias
    par['IFyy'] = inertia.compound_pendulum_inertia(mp['mF'], mp['g'],
                                                    mp['lF'], mp['TcF1'])
    par['IRyy'] = inertia.compound_pendulum_inertia(mp['mR'], mp['g'],
                                                    mp['lR'], mp['TcR1'])
    try:
        # we measured the inertia of the front wheel with the flywheel inside
        iFlywheelPlusFwheel = inertia.compound_pendulum_inertia(
            mp['mD'], mp['g'], mp['lF'], mp['TcD1'])
        par['IDyy'] = iFlywheelPlusFwheel - par['IFyy']
    except KeyError:
        pass

    # calculate the y inertias for the frame and fork
    lB = (par['xB']**2 + (par['zB'] + par['rR'])**2)**(0.5)
    par['IByy'] = inertia.compound_pendulum_inertia(mp['mB'], mp['g'], lB,
                                                    mp['TcB1'])

    if forkIsSplit:
        # fork
        lS = ((par['xS'] - par['w'])**2 +
              (par['zS'] + par['rF'])**2)**(0.5)
        par['ISyy'] = inertia.compound_pendulum_inertia(mp['mS'], mp['g'], lS,
                                                        mp['TcS1'])
        # handlebar
        l1, l2 = geometry.calculate_l1_l2(mp['h6'], mp['h7'], mp['d5'],
                                          mp['d6'], mp['l'])
        u1, u2 = geometry.fwheel_to_handlebar_ref(par['lam'], l1, l2)
        lG = ((par['xG'] - par['w'] + u1)**2 +
              (par['zG'] + par['rF'] + u2)**2)**(.5)
        par['IGyy'] = inertia.compound_pendulum_inertia(mp['mG'], mp['g'], lG,
                                                        mp['TcG1'])
    else:
        lH = ((par['xH'] - par['w'])**2 +
              (par['zH'] + par['rF'])**2)**(0.5)
        par['IHyy'] = inertia.compound_pendulum_inertia(mp['mH'], mp['g'], lH,
                                                        mp['TcH1'])

    # calculate the stiffness of the torsional pendulum
    IPxx, IPyy, IPzz = inertia.tube_inertia(mp['lP'], mp['mP'],
                                            mp['dP'] / 2., 0.)
    torStiff = inertia.torsional_pendulum_stiffness(IPyy, mp['TtP1'])

    # calculate the wheel x/z inertias
    par['IFxx'] = inertia.tor_inertia(torStiff, mp['TtF1'])
    par['IRxx'] = inertia.tor_inertia(torStiff, mp['TtR1'])
    try:
        par['IDxx'] = inertia.tor_inertia(torStiff, mp['TtD1']) - par['IFxx']
    except KeyError:
        pass

    pendulumInertias = {}

    # calculate the in plane moments of inertia
    for part, slopeSet in slopes.items():
        # the number of orientations for this part
        numOrien = len(slopeSet)
        # intialize arrays to store the inertia values and orientation angles
        penInertia = np.zeros(numOrien, dtype=object)
        beta = np.array(betas[part])
        # fill arrays of the inertias
        for i in range(numOrien):
            penInertia[i] = inertia.tor_inertia(torStiff,
                                                mp['Tt' + part + str(i + 1)])
        # store these inertias
        pendulumInertias[part] = list(penInertia)
        inert = inertia.inertia_components(penInertia, beta)
        for i, axis in enumerate(['xx', 'xz', 'zz']):
            par['I' + part + axis] = inert[i]

    if forkIsSplit:
        # combine the moments of inertia to find the total handlebar/fork MoI
        IG = inertia.part_inertia_tensor(par, 'G')
        IS = inertia.part_inertia_tensor(par, 'S')
        # columns are parts, rows = x, y, z
        coordinates = np.array([[par['xG'], par['xS']],
                                [0., 0.],
                                [par['zG'], par['zS']]])
        masses = np.array([par['mG'], par['mS']])
        par['mH'], cH = com.total_com(coordinates, masses)
        par['xH'] = cH[0]
        par['zH'] = cH[2]
        dG = np.array([par['xG'] - par['xH'], 0., par['zG'] - par['zH']])
        dS = np.array([par['xS'] - par['xH'], 0., par['zS'] - par['zH']])
        IH = (inertia.parallel_axis(IG, par['mG'], dG) +
              inertia.parallel_axis(IS, par['mS'], dS))
        par['IHxx'] = IH[0, 0]
        par['IHxz'] = IH[0, 2]
        par['IHyy'] = IH[1, 1]
        par['IHzz'] = IH[2, 2]

    # package the extra information that is useful outside this function
    extras = {'slopes': slopes,
              'intercepts': intercepts,
              'betas': betas,
              'pendulumInertias': pendulumInertias}

    return par, extras


def is_fork_split(mp):
    '''Returns true if the fork was split into two parts and false if not.

    Parameters
    ----------
    mp : dictionary
        The measured data.

    Returns
    -------
    forkIsSplit : boolean

    '''
    # this isn't that robust, for example if you had an S and no G then this
    # wouldn't catch it
    forkIsSplit = False
    for key in mp.keys():
        # if there is an 'S' then the fork is split in two parts
        if key[:1] == 'S' or key[1:2] == 'S':
            forkIsSplit = True

    return forkIsSplit



import os
import time

try:
    import dash
except ImportError as e:
    msg = ('Dash not installed, make sure to install the optional '
           'dependencies for the web app.')
    raise ImportError(msg) from e

try:
    import dash_bootstrap_components as dbc
except ImportError as e:
    msg = ('dash_bootstrap_components not installed, make sure to install the '
           'optional dependencies for the web app.')
    raise ImportError(msg) from e

try:
    import pandas
except ImportError as e:
    msg = ('pandas not installed, make sure to install the optional '
           'dependencies for the web app.')
    raise ImportError(msg) from e
else:
    del pandas

from dash import dash_table as tbl
from dash import dcc
from dash import html
from dash.dependencies import Input, Output
import numpy as np

# TODO : should this be a relative import?
import bicycleparameters as bp

PATH_TO_THIS_FILE = os.path.dirname(os.path.abspath(__file__))
PATH_TO_APP_DATA = os.path.join(PATH_TO_THIS_FILE, 'app-data')
PATH_TO_ASSETS = os.path.join(PATH_TO_THIS_FILE, 'assets')

# list of all the whipple-carvallo parameters

par_list = ['rF', 'mF', 'IFxx', 'IFyy', 'rR', 'mR', 'IRxx', 'IRyy',
            'w', 'c', 'lam', 'g',
            'xB', 'zB', 'mB',
            'IBxx', 'IByy', 'IBzz', 'IBxz',
            'xH', 'zH', 'mH',
            'IHxx', 'IHyy', 'IHzz', 'IHxz']

# list of bicycle models in the dropdown menu

OPTIONS = ['Benchmark',
           'Browser',
           'Browserins',
           'Crescendo',
           'Fisher',
           'Pista',
           'Rigid',
           'Silver',
           'Yellow',
           'Yellowrev']

OPT_LABELS = [
    'Benchmark',
    'Browser (no rider)',
    'Browserins (no rider)',
    'Crescendo (no rider)',
    'Fisher (no rider)',
    'Pista (no rider)',
    'Rigid (no rider)',
    'Silver (no rider)',
    'Yellow (no rider)',
    'Yellowrev (no rider)',
]

WHEEL_COLUMNS = [{'name': '', 'id': 'label', 'type': 'text', 'editable': False},
                 {'name': 'Front Wheel [F]', 'id': 'fW', 'type': 'numeric'},
                 {'name': 'Rear Wheel [R]', 'id': 'rW', 'type': 'numeric'}]

WHEEL_LABELS = ['Radius [m]:',
                'Mass [kg]:',
                'Moment Ixx [kg*m²]:',
                'Moment Iyy [kg*m²]:']

FRAME_COLUMNS = [{'name': '', 'id': 'label', 'type': 'text', 'editable': False},
                 {'name': 'Rear Body [B]', 'id': 'rB', 'type': 'numeric'},
                 {'name': 'Front Assembly [H]', 'id': 'fA', 'type': 'numeric'}]

FRAME_LABELS = ['Center of Mass X [m]:',
                'Center of Mass Z [m]:',
                'Total Mass [kg]:',
                'Moment Ixx [kg*m²]:',
                'Moment Iyy [kg*m²]:',
                'Moment Izz [kg*m²]:',
                'Moment Ixz [kg*m²]:']

GENERAL_COLUMNS = [{'name': '', 'id': 'label', 'type': 'text', 'editable': False},
                   {'name': 'Parameters', 'id': 'data', 'type': 'numeric'}]

GENERAL_LABELS = ['Wheel Base [m]:',
                  'Trail [m]:',
                  'Steer Axis Tilt [degrees]:',
                  'Gravity [N/kg]:']

app = dash.Dash(__name__, external_stylesheets=[dbc.themes.COSMO])
app.title = 'Bicycle Dynamics Analysis App'
server = app.server  # needed for heroku or render


# app.layout defines the visual GUI elements on the website

app.layout = html.Div([
    dbc.Container(fluid=False,
                  children=[
                      dbc.Row([dbc.Col(dbc.NavbarSimple(brand='Bicycle Dynamics Analysis App',
                                                        dark=True,
                                                        color='primary'),
                                       width='auto',
                                       className='my-2')],
                                       justify='center'),
                      dbc.Row([dbc.Col(dcc.Loading(id='geometry-load',
                                                   type='dot',
                                                   children=[html.Div(id='geometry-bin',
                                                                      children=[html.Div([dcc.Graph(
                                                                                         id='geometry-plot',
                                                                                         className='img-fluid')])])]),
                                       lg=5,
                                       width=12),
                               dbc.Col(dcc.Loading(id='eigen-load',
                                                   type='dot',
                                                   children=[html.Div(id='eigen-bin',
                                                                      children=[html.Div([dcc.Graph(
                                                                                         id='eigen-plot',
                                                                                         className='img-fluid')])])]),
                                       lg=5,
                                       width=12),
                               dbc.Col(children=[html.H5('Choose a Parameter Set:',
                                                         className='centered'),
                                                 dcc.Dropdown(id='bike-dropdown',
                                                              value='Benchmark',
                                                              options=[{'label': k, 'value': v} for k, v in zip(OPT_LABELS, OPTIONS)],
                                                              style={'color': 'black'}),
                                                 dcc.Checklist(id='geometry-checklist',
                                                               options=[{'label': 'Show Centers of Mass ', 'value': 'centers'},
                                                                        {'label': 'Show Inertia Ellipsoids ', 'value': 'ellipse'},
                                                                        {'label': 'Show Self-stability region', 'value': 'stability'}],
                                                               value=['centers']),
                                                 dbc.Button('Reset Table',
                                                            id='reset-button',
                                                            color='primary',
                                                            size='lg',
                                                            n_clicks=0)],
                                       lg=2)],
                              className="my-2"),
                      dbc.Row([dbc.Col(tbl.DataTable(id='frame-table',
                                                     columns=FRAME_COLUMNS,
                                                     data=[],
                                                     style_cell={
                                                          'backgroundColor': 'white',
                                                          'color': 'black',
                                                          # 'border': '1px solid white',
                                                          'whiteSpace': 'normal',
                                                          'height': 'auto'
                                                      },
                                                      style_header={
                                                           'textAlign': 'center',
                                                           'backgroundColor': 'rgba(39,128,227,0.15)'
                                                      },
                                                      style_data_conditional=[{
                                                             "if": {"state": "selected"},
                                                              'backgroundColor': 'rgba(39,128,227,0.15)',
                                                              'color': 'black',
                                                              "border": "1px solid black",
                                                      },],
                                                      editable=True),
                                       lg=4),
                               dbc.Col(children=[html.H5('Set the EigenValue Speed Range [m/s]:',
                                                         className='centered'),
                                                 dcc.RangeSlider(id='range-slider',
                                                                 min=-45,
                                                                 max=45,
                                                                 step=5,
                                                                 value=[0, 10],
                                                                 marks={-40: {'label': '-40'},
                                                                        -30: {'label': '-30'},
                                                                        -20: {'label': '-20'},
                                                                        -10: {'label': '-10'},
                                                                        0: {'label': '0', },
                                                                        10: {'label': '10'},
                                                                        20: {'label': '20'},
                                                                        30: {'label': '30'},
                                                                        40: {'label': '40'}},
                                                                 allowCross=False),
                                                 dbc.Row([dbc.Col(tbl.DataTable(id='wheel-table',
                                                                                columns=WHEEL_COLUMNS,
                                                                                data=[],
                                                                                style_cell={
                                                                                    'whiteSpace': 'normal',
                                                                                    'height': 'auto'
                                                                                },
                                                                                style_header={
                                                                                    'textAlign': 'center',
                                                                                    'backgroundColor': 'rgba(39,128,227,0.15)'
                                                                                },
                                                                                style_data_conditional=[
                                                                                    {
                                                                                        "if": {"state": "selected"},
                                                                                        "backgroundColor": 'rgba(39,128,227,0.15)',
                                                                                        'color': 'black',
                                                                                        "border": "1px solid black",
                                                                                    },
                                                                                ],
                                                                                editable=True)),
                                                          dbc.Col(tbl.DataTable(id='general-table',
                                                                                columns=GENERAL_COLUMNS,
                                                                                data=[],
                                                                                style_cell={
                                                                                    'whiteSpace': 'normal',
                                                                                    'height': 'auto'
                                                                                },
                                                                                style_header={
                                                                                    'textAlign': 'center',
                                                                                    'backgroundColor': 'rgba(39,128,227,0.15)'
                                                                                },
                                                                                style_data_conditional=[
                                                                                    {
                                                                                        "if": {"state": "selected"},
                                                                                        "backgroundColor": 'rgba(39,128,227,0.15)',
                                                                                        'color': 'black',
                                                                                        "border": "1px solid black",
                                                                                    },
                                                                                ],
                                                                                editable=True))
                                                          ])
                                                 ],
                                       lg=8,
                                       align='end')],
                              className="my-2"),
                      dbc.Row([dbc.Col(dcc.Markdown(open(os.path.join(PATH_TO_ASSETS, 'app-explanation.md')).read()),
                                       width='auto',
                                       className='my-2')
                               ])
                  ])])


# create a set of Benchmark parameters with uncertainties removed

def new_par(bike_name):
    bike = bp.Bicycle(bike_name, pathToData=PATH_TO_APP_DATA)
    par = bike.parameters['Benchmark']
    parPure = bp.io.remove_uncertainties(par)
    return parPure


# populates all data tables with bicycleparameters data

@app.callback([Output('wheel-table', 'data'),
               Output('frame-table', 'data'),
               Output('general-table', 'data')],
              [Input('bike-dropdown', 'value'),
               Input('reset-button', 'n_clicks')])
def populate_wheel_data(value, n_clicks):

    # generates data for wheel-table
    wheelPure = new_par(value)
    wheelData = []
    wheelLabels = []
    fW = []
    rW = []
    for i in WHEEL_LABELS:
        wheelLabels.append({'label': i})
    for i in range(8):
        if i < 4:
            fW.append({'fW': wheelPure.get(par_list[i])})
        else:
            rW.append({'rW': wheelPure.get(par_list[i])})
    for c, d, e in zip(wheelLabels, fW, rW):
        zipped = {}
        zipped.update(c)
        zipped.update(d)
        zipped.update(e)
        wheelData.append(zipped)

    # generates data for frame-table
    framePure = new_par(value)
    frameData = []
    frameLabels = []
    rB = []
    fA = []
    for i in FRAME_LABELS:
        frameLabels.append({'label': i})
    for i in range(12, len(par_list)):
        if i < 19:
            rB.append({'rB': framePure.get(par_list[i])})
        else:
            fA.append({'fA': framePure.get(par_list[i])})
    for c, d, e in zip(frameLabels, rB, fA):
        zipped = {}
        zipped.update(c)
        zipped.update(d)
        zipped.update(e)
        frameData.append(zipped)

    # generates data for general-table
    genPure = new_par(value)
    genData = []
    genLabels = []
    data = []
    for i in GENERAL_LABELS:
        genLabels.append({'label': i})
    for i in range(8, 12):
        data.append({'data': genPure.get(par_list[i])})

    # converts radians to degrees for display in the table
    lamD = data[2]
    radians = lamD.get('data')
    degrees = np.rad2deg(radians)
    data[2] = {'data': int(degrees)}

    for c, d in zip(genLabels, data):
        zipped = {}
        zipped.update(c)
        zipped.update(d)
        genData.append(zipped)

    return wheelData, frameData, genData


# updates geometry-plot & eigen-plot path with Dropdown value or edited
# DataTable values

@app.callback([Output('geometry-plot', 'figure'),
               Output('eigen-plot', 'figure')],
              [Input('bike-dropdown', 'value'),
               Input('wheel-table', 'data'),
               Input('frame-table', 'data'),
               Input('general-table', 'data'),
               Input('geometry-checklist', 'value'),
               Input('range-slider', 'value')])
def plot_update(value, wheel, frame, general, options, slider):

    # accesses Input properties to avoid redundancies
    ctx = dash.callback_context
    wheelData = ctx.inputs.get('wheel-table.data')
    frameData = ctx.inputs.get('frame-table.data')
    genData = ctx.inputs.get('general-table.data')
    checklistData = ctx.inputs.get('geometry-checklist.value')
    rangeSliderData = ctx.inputs.get('range-slider.value')

    # construct flags for selected values of the geometry plot display options
    mass_boolean = 'centers' in checklistData
    ellipse_boolean = 'ellipse' in checklistData
    stability_option = 'stability' in checklistData

    # sets the speed range for eigen-plot based on range-slider
    minBound = rangeSliderData[0]
    maxBound = rangeSliderData[1]
    steps = (maxBound-minBound)/0.1
    speeds = np.linspace(minBound, maxBound, num=int(steps))

    # create Bike using default data based on dropdown menu value
    Bike = bp.Bicycle(value, pathToData=PATH_TO_APP_DATA)

    # convert steer axis tilt into radians if recieving values from datatable
    # edits
    if ctx.triggered[0].get('prop_id') != 'bike-dropdown.value':

        # convert to radians
        degrees = float(genData[2].get('data'))
        radians = np.deg2rad(degrees)
        genData[2]['data'] = radians

        # creates an alternating list of [parameter,value] from table data
        newP = []
        for p in range(8):
            if p < 4:
                newP.extend([par_list[p], wheelData[p].get('fW')])
            else:
                newP.extend([par_list[p], wheelData[p-4].get('rW')])
        for p in range(12, len(par_list)):
            if p < 19:
                newP.extend([par_list[p], frameData[p-12].get('rB')])
            else:
                newP.extend([par_list[p], frameData[p-19].get('fA')])
        for p in range(8, 12):
            newP.extend([par_list[p], genData[p-8].get('data')])

        # inserts user edited data into the default Bike created earlier
        for i in range(0, len(newP), 2):
            Bike.parameters['Benchmark'][newP[i]] = newP[i+1]

    # create geometry-plot with plotly

    geo_plot = Bike._plot_bicycle_geometry_plotly(
        show=False, centerOfMass=mass_boolean, inertiaEllipse=ellipse_boolean)
    # Create eigenvalues-plot with plotly
    eigen_plot = Bike._plot_eigenvalues_vs_speed_plotly(
        speeds, show=False, stability_region=stability_option)

    return eigen_plot, geo_plot


# sets loading notification for the geometry plot
@app.callback(Output("geometry-load", "children"),
              [Input("geometry-bin", "children")])
def input_triggers_spinner1(value):
    time.sleep(1)
    return value


# sets loading notification for the eigenvalue plot
@app.callback(Output("eigen-load", "children"),
              [Input("eigen-bin", "children")])
def input_triggers_spinner2(value):
    time.sleep(1)
    return value


# omit the `dev_tools_ui` parameter to display debug info in the browser rather
# than in the terminal
if __name__ == '__main__':
    app.run_server(debug=True, dev_tools_ui=False)



#!/usr/bin/env python

from setuptools import setup, find_packages

exec(open('bicycleparameters/version.py').read())

setup(
    name='BicycleParameters',
    version=__version__,
    author='Jason K. Moore',
    author_email='moorepants@gmail.com',
    packages=find_packages(),
    url='http://pypi.python.org/pypi/BicycleParameters',
    license='LICENSE.txt',
    description='Generates and manipulates the physical parameters of a bicycle.',
    long_description=open('README.rst').read(),
    project_urls={
        'Web Application': 'https://bicycle-dynamics.onrender.com',
        'Documentation': 'http://bicycleparameters.readthedocs.io',
        'Source Code': 'https://github.com/moorepants/BicycleParameters',
        'Issue Tracker': 'https://github.com/moorepants/BicycleParameters/issues',
    },
    include_package_data=True,  # includes things in MANIFEST.in
    install_requires=[
        'DynamicistToolKit>=0.5.3',
        'matplotlib>=3.5.1',
        'numpy>=1.21.5',
        'pyyaml>=5.4.1',
        'scipy>=1.8.0',
        'uncertainties>=3.1.5',
        'yeadon>=1.3.0',
    ],
    python_requires='>=3.8',
    extras_require={
        'doc': [
            'sphinx>=4.3.2',
            'numpydoc>=1.2'
        ],
        'app': [
            'dash>=2',
            'dash-bootstrap-components',
            'pandas>=1.3.5'
        ],
    },
    entry_points={'console_scripts':
                  ['bicycleparameters = bicycleparameters.app:app.run_server']},
    classifiers=['Programming Language :: Python',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: 3.11',
                 'Programming Language :: Python :: 3.12',
                 'Operating System :: OS Independent',
                 'Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: BSD License',
                 'Natural Language :: English',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Physics']
)



import numpy as np

from .com import total_com
from .inertia import parallel_axis


def _rotate_inertia_about_y(inertia, angle):
    """Returns inertia tensor rotated through an angle about the Y axis.

    Parameters
    ==========
    inertia : array_like, shape(3,3)
        An inertia tensor of a single rigid body defined relative to an XYZ
        Cartesian coordinate system.
    angle : float
        Angle in radians about the positive Y axis of which to rotate the
        coordinate system.

    Returns
    =======
    ndarray, shape(3,3)
        An inertia tensor of a single rigid body expressed relative to a
        rotated Cartesian coordinate system.

    """
    ca = np.cos(angle)
    sa = np.sin(angle)
    C = np.array([[ca, 0.0, -sa],
                  [0.0, 1., 0.0],
                  [sa, 0.0, ca]])
    return C @ inertia @ C.T


def _sum_central_inertias(m1, com1, I1, m2, com2, I2):
    # SUM_CENTRAL_INERTIAS - Returns the combined inertia of two bodies.
    #
    # Syntax : I = sum_central_inertias(m1, com1, I1, m2, com2, I2)
    #
    # Inputs:
    #   m1 - 1x1 double, mass of the first body
    #   com1 - 3x1 double, center of mass of the first body
    #   I1 - 3x3 double, central inertia tensor of the first body
    #   m2 - 1x1 double, mass of the second body
    #   com2 - 3x1 double, center of mass of the second body
    #   I2 - 3x3 double, central inertia tensor of the second body
    # Outputs:
    #   I - combined central inertia tensor of both bodies
    m, com = total_com(np.vstack([com1, com2]).T, [m1, m2])
    return parallel_axis(I1, m1, com-com1) + parallel_axis(I2, m2, com-com2)


def _convert_principal_to_benchmark(principal_par):
    # CONVERT_PRINCIPAL_TO_BENCHMARK - Returns a structure containing the
    # benchmark bicycle parameters as defined in Meijaard2007 which are
    # converted from the principal parameters as defined in MooreHubbard2019.
    #
    # Syntax: benchmark_par = convert_principal_to_benchmark(principal_par)
    #
    # Inputs:
    #   principal_par - Structure containing parameter names mapped to doubles.
    # Outputs:
    #   benchmark_par - Structure containing parameter names mapped to doubles.

    p = principal_par

    b = {}

    ###########################################################################
    # primary geometry, gravity, and speed
    ###########################################################################
    b["w"] = p["w"]
    b["c"] = p["c"]
    b["lam"] = p["lam"]
    b["g"] = p["g"]
    b["v"] = p["v"]

    ###########################################################################
    # rear frame [B]
    ###########################################################################
    coordinates = np.array([[p["xD"], p["xP"]],
                            [p["yD"], p["yP"]],
                            [p["zD"], p["zP"]]])
    masses = np.array([p["mD"], p["mP"]])
    b["mB"], (b["xB"], b["yB"], b["zB"]) = total_com(coordinates, masses)

    # symmetry is assumed about the XZ plane
    ID_principal = p["mD"]*np.diag([p["kDaa"]**2, p["kDyy"]**2, p["kDbb"]**2])
    IDxyz = _rotate_inertia_about_y(ID_principal, -p["alphaD"])

    # person
    IP_principal = p["mP"]*np.diag([p["kPaa"]**2, p["kPyy"]**2, p["kPbb"]**2])
    IPxyz = _rotate_inertia_about_y(IP_principal, -p["alphaP"])

    # combined rear frame and person
    comD = np.array([p["xD"], p["yD"], p["zD"]])
    comP = np.array([p["xP"], p["yP"], p["zP"]])
    IBxyz = _sum_central_inertias(p["mD"], comD, IDxyz, p["mP"], comP, IPxyz)
    b["IBxx"] = IBxyz[0, 0]
    b["IByy"] = IBxyz[1, 1]
    b["IBzz"] = IBxyz[2, 2]
    b["IBxz"] = IBxyz[0, 2]

    ###########################################################################
    # front frame [H]
    ###########################################################################
    b["mH"] = p["mH"]
    b["xH"] = p["xH"]
    b["yH"] = p["yH"]
    b["zH"] = p["zH"]

    # symmetry is assumed about the XZ plane
    IH123 = p["mH"]*np.diag([p["kHaa"]**2, p["kHyy"]**2, p["kHbb"]**2])
    IH = _rotate_inertia_about_y(IH123, -p["alphaH"])
    b["IHxx"] = IH[0, 0]
    b["IHyy"] = IH[1, 1]
    b["IHzz"] = IH[2, 2]
    b["IHxz"] = IH[0, 2]

    ###########################################################################
    # rear wheel [R]
    ###########################################################################
    b["rR"] = p["rR"]
    b["mR"] = p["mR"]
    # wheel is symmetric about XY, YZ, and XZ planes, thus no products of
    # inertia, wheel is ring or disc like, thus Ixx=Izz and Iyy > Ixx
    b["IRxx"] = p["mR"] * p["kRaa"]**2
    b["IRyy"] = p["mR"] * p["kRyy"]**2
    b["IRzz"] = p["mR"] * p["kRaa"]**2

    ###########################################################################
    # front wheel [F]
    ###########################################################################
    b["rF"] = p["rF"]
    b["mF"] = p["mF"]
    # wheel is symmetric about XY, YZ, and XZ planes, thus no prodcuts of
    # inertia, wheel is always ring or disc like, thus Ixx=Izz and Iyy > Ixx
    b["IFxx"] = p["mF"] * p["kFaa"]**2
    b["IFyy"] = p["mF"] * p["kFyy"]**2
    b["IFzz"] = p["mF"] * p["kFaa"]**2

    # return the benchmark parameters
    return b



import os

import yaml
import numpy as np
import matplotlib.pyplot as plt

from ..parameter_sets import Meijaard2007ParameterSet, Moore2019ParameterSet

PARDIR_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'parameter_sets')


def test_Meijaard2007ParameterSet(plot=False):

    filepath = os.path.join(PARDIR_PATH, 'benchmark-benchmark.yml')
    with open(filepath, 'r') as f:
        benchmark_par = yaml.load(f, Loader=yaml.FullLoader)['values']

    pset = Meijaard2007ParameterSet(benchmark_par, True)

    pset2 = Meijaard2007ParameterSet._from_yaml(filepath)

    assert pset.parameters == pset2.parameters
    assert pset2.includes_rider

    assert pset.includes_rider is True
    assert pset.parameters['v'] == 5.0

    expected_IH = np.array([[0.05892, 0.0, -0.00756],
                            [0.0, 0.06, 0.0],
                            [-0.00756, 0.0, 0.00708]])
    np.testing.assert_allclose(pset.form_inertia_tensor('H'),
                               expected_IH)

    expected_comB = np.array([0.3, 0.0, -0.9])
    expected_comF = np.array([1.02, 0.0, -0.35])
    expected_comH = np.array([0.9, 0.0, -0.7])
    expected_comR = np.array([0.0, 0.0, -0.3])

    np.testing.assert_allclose(expected_comB,
                               pset.form_mass_center_vector('B'))
    np.testing.assert_allclose(expected_comF,
                               pset.form_mass_center_vector('F'))
    np.testing.assert_allclose(expected_comH,
                               pset.form_mass_center_vector('H'))
    np.testing.assert_allclose(expected_comR,
                               pset.form_mass_center_vector('R'))

    np.testing.assert_allclose(pset.form_mass_center_vector('H'),
                               pset.mass_center_of('H'))

    com_total = pset.mass_center_of('F', 'R')

    x = (1.02*3.0 + 0.0*2.0)/(3.0 + 2.0)
    y = 0.0
    z = (-0.35*3.0 + -0.3*2.0)/(3.0 + 2.0)

    np.testing.assert_allclose(com_total, np.array([x, y, z]))

    if plot:
        ax = pset.plot_geometry()
        ax = pset.plot_principal_radii_of_gyration(ax=ax)
        ax = pset.plot_principal_inertia_ellipsoids(ax=ax)
        ax = pset.plot_mass_centers(ax=ax)
        plt.show()


def test_Moore2019ParameterSet(plot=False):

    with open(os.path.join(PARDIR_PATH,
                           'principal-browserjason.yml'), 'r') as f:
        principal_par = yaml.load(f, Loader=yaml.FullLoader)['values']

    pset = Moore2019ParameterSet(principal_par)

    ax = pset.plot_geometry()
    ax = pset.plot_person_diamond(ax=ax)
    ax = pset.plot_principal_radii_of_gyration(ax=ax)
    ax = pset.plot_body_mass_center('D', ax=ax)
    ax = pset.plot_body_mass_center('F', ax=ax)
    ax = pset.plot_body_mass_center('H', ax=ax)
    ax = pset.plot_body_mass_center('P', ax=ax)
    ax = pset.plot_body_mass_center('R', ax=ax)
    ax = pset.plot_body_principal_inertia_ellipsoid('D', ax=ax)
    ax = pset.plot_body_principal_inertia_ellipsoid('P', ax=ax)
    ax = pset.plot_body_principal_inertia_ellipsoid('H', ax=ax)

    if plot:
        plt.show()


def test_conversion(plot=False):
    with open(os.path.join(PARDIR_PATH,
                           'principal-browserjason.yml'), 'r') as f:
        par_dict = yaml.load(f, Loader=yaml.FullLoader)['values']
    pset = Moore2019ParameterSet(par_dict)
    pset.plot_all()
    bench_pset = pset.to_parameterization('Meijaard2007')
    bench_pset.plot_all()
    if plot:
        plt.show()



from math import ceil


class Table():
    """A class for generating tables of the measurment and parameter data
    associated with a bicycle. """

    def __init__(self, source, latex, bicycles):
        """Sets the basic attributes of the table.

        Parameters
        ----------
        source : string
            One of the parameter types: `Measured` or `Benchmark` for now.
        latex : boolean
            If true, the variable names will be formatted with LaTeX.
        bicycles : tuple or list of Bicycles
            Bicycle objects of which their parameters should appear in the
            generated table. The order of the bicycles determines the order in
            the table.

        """
        self.source = source
        self.bicycles = bicycles
        self.latex = latex
        # go ahead and calculate the base data, which sets allVariables and
        # tableData
        self.generate_variable_list()
        self.generate_table_data()

    def generate_variable_list(self):
        # generate a complete list of the variables
        allVariables = []
        try:
            for bicycle in self.bicycles:
                allVariables += bicycle.parameters[self.source].keys()
        except TypeError:
            allVariables += self.bicycle.parameters[self.source].keys()
        # remove duplicates and sort
        self.allVariables = sorted(list(set(allVariables)), key=lambda x:
                                   x.lower())

    def generate_table_data(self):
        """Generates a list of data for a table."""
        table = []
        for var in self.allVariables:
            # add a new line
            table.append([])

            if self.latex:
                table[-1].append(to_latex(var))
            else:
                table[-1].append(var)

            for bicycle in self.bicycles:
                try:
                    val, sig = uround(bicycle.parameters[self.source][var]).split('+/-')
                except ValueError:
                    val = str(bicycle.parameters[self.source][var])
                    sig = 'NA'
                except KeyError:
                    val = 'NA'
                    sig = 'NA'
                table[-1] += [val, sig]

        self.tableData = table

    def create_rst_table(self, fileName=None):
        """Returns a reStructuredText version of the table.

        Parameters
        ----------
        fileName : string
            If a path to a file is given, the table will be written to that
            file.

        Returns
        -------
        rstTable : string
            reStructuredText version of the table.

        """

        table = self.tableData

        # add the math directive if using latex
        if self.latex:
            for i, row in enumerate(table):
                self.tableData[i][0] = ':math:`' + row[0] + '`'

        # add a sub header
        table.insert(0, ['Variable'])
        for i, bicycle in enumerate(self.bicycles):
            if self.latex:
                table[0] += [':math:`v`', r':math:`\sigma`']
            else:
                table[0] += ['v', 'sigma']

        # find the longest string in each column
        largest = [0] # the top left is empty
        for bicycle in self.bicycles:
            l = int(ceil(len(bicycle.bicycleName) / 2.0))
            largest += [l, l]
        for row in table:
            colSize = [len(string) for string in row]
            for i, pair in enumerate(zip(colSize, largest)):
                if pair[0] > pair[1]:
                    largest[i] = pair[0]

        # build the rst table
        rstTable = '+' + '-' * (largest[0] + 2)

        for i, bicycle in enumerate(self.bicycles):
            rstTable += '+' + '-' * (largest[2 * i + 1] + largest[2 * i + 2] + 5)

        rstTable += '+\n|' + ' ' * (largest[0] + 2)

        for i, bicycle in enumerate(self.bicycles):
            rstTable += '| ' + bicycle.bicycleName + ' ' * (largest[2 * i + 1] +
                    largest[2 * i + 2] + 4 - len(bicycle.bicycleName))
        rstTable += '|\n'

        for j, row in enumerate(table):
            if j == 0:
                dash = '='
            else:
                dash = '-'

            line = ''
            for i in range(len(row)):
                line += '+' + dash * (largest[i] + 2)

            line += '+\n|'
            for i, item in enumerate(row):
                line += ' ' + item + ' ' * (largest[i] - len(item)) + ' |'
            line += '\n'
            rstTable += line

        for num in largest:
           rstTable += '+' + dash * (num + 2)
        rstTable += '+'

        if fileName is not None:
            f = open(fileName, 'w')
            f.write(rstTable)
            f.close()

        return rstTable


def to_latex(var):
    """Returns a latex representation for a given variable string name.

    Parameters
    ----------
    var : string
        One of the variable names used in the bicycleparameters package.

    Returns
    -------
    latex : string
        A string formatting for pretty LaTeX math print.

    """

    latexMap = {'f': 'f',
                'w': 'w',
                'gamma': r'\gamma',
                'g': 'g',
                'lcs': 'l_{cs}',
                'hbb': 'h_{bb}',
                'lsp': 'l_{sp}',
                'lst': 'l_{st}',
                'lamst': r'\lambda_{st}',
                'whb': 'w_{hb}',
                'LhbF': 'l_{hbF}',
                'LhbR': 'l_{hbR}',
                'd': 'd',
                'l': 'l',
                'c': 'c',
                'lam': r'\lambda',
                'xcl': 'x_{cl}',
                'zcl': 'z_{cl}',
                'ds1': 'd_{s1}',
                'ds3': 'd_{s3}'}
    try:
        latex = latexMap[var]
    except KeyError:
        if var.startswith('alpha'):
            latex = r'\alpha_{' + var[-2:] + '}'
        elif var.startswith('a') and len(var) == 3:
            latex = 'a_{' + var[-2:] + '}'
        elif var.startswith('T'):
            latex = 'T^' + var[1] + '_{' + var[-2:] + '}'
        elif len(var) == 2:
            latex = var[0] + '_' + var[1]
        elif var.startswith('I'):
            latex = var[0] + '_{' + var[1:] + '}'
        else:
            raise

    return latex


def uround(value):
    '''Returns a string representation of a value with an uncertainity which
    has been rounded to significant digits based on the uncertainty value.

    Parameters
    ----------
    value : ufloat
        A single ufloat.

    Returns
    -------
    s : string
        A rounded string representation of `value`.

    2.4563752289999+/-0.0003797273827

    becomes

    2.4564+/-0.0004

    This probably doesn't work for weird cases like large uncertainties.

    '''
    try:
    # grab the nominal value and the uncertainty
        nom = value.nominal_value
    except AttributeError:
        s = str(value)
    else:
        uncert = value.std_dev
        if abs(nom) < 1e-15:
            s = '0.0+/-0.0'
        else:
            # convert the uncertainty to a string
            s = '%.14f' % uncert
            # find the first non-zero character
            for j, number in enumerate(s):
                if number == '0' or number == '.':
                    pass
                else:
                    digit = j
                    break
            newUncert = round(uncert, digit - 1)
            newUncertStr = ('%.' + str(digit - 1) + 'f') % newUncert
            newNom = round(nom, len(newUncertStr) - 2)
            newNomStr = ('%.' + str(digit - 1) + 'f') % newNom
            diff = len(newUncertStr) - len(newNomStr)
            if diff > 0:
                s = newNomStr + int(diff) * '0' + '+/-' + newUncertStr
            else:
                s = newNomStr + '+/-' + newUncertStr
    return s
